"""Nexus grid file class for loading in a structured grid file and extracting the grid properties."""
from __future__ import annotations

import copy
import pandas as pd
from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING

from ResSimpy.Grid import Grid, VariableEntry
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.structured_grid_operations import StructuredGridOperations
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.Nexus.array_function_operations as afo

from resqpy.olio.read_nexus_fault import load_nexus_fault_mult_table_from_list

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


@dataclass
class PropertyToLoad:
    token: str
    modifiers: list[str]
    property: VariableEntry


@dataclass(kw_only=True)
class NexusGrid(Grid):
    __array_functions_list: Optional[list[list[str]]] = None
    __array_functions_df: Optional[pd.DataFrame] = None
    __array_functions_loaded: bool = False
    __grid_file_contents: Optional[list[str]] = None
    __grid_file_nested: Optional[list[str]] = None
    __faults_df: Optional[pd.DataFrame] = None
    __grid_faults_loaded: bool = False
    __grid_properties_loaded: bool = False
    __grid_nexus_file: Optional[NexusFile] = None

    def __init__(self, grid_nexus_file: Optional[NexusFile] = None) -> None:
        super().__init__()
        self.__array_functions_list: Optional[list[str]] = None
        self.__array_functions_df: Optional[pd.DataFrame] = None
        self.__array_functions_loaded: bool = False
        self.__grid_file_contents: Optional[list[str]] = None if grid_nexus_file is None else \
            grid_nexus_file.get_flat_list_str_file_including_includes
        self.__grid_file_nested: Optional[list[str]] = None if grid_nexus_file is None else \
            grid_nexus_file.file_content_as_list
        self.__faults_df: Optional[pd.DataFrame] = None
        self.__grid_faults_loaded: bool = False
        self.__grid_properties_loaded: bool = False
        self.__grid_nexus_file: Optional[NexusFile] = grid_nexus_file

    def __wrap(self, value):
        if isinstance(value, tuple | list | set | frozenset):
            return type(value)([self.__wrap(v) for v in value])
        else:
            return value

    def update_properties_from_dict(self, data: dict[str, int | VariableEntry]) -> None:
        """Allows you to update properties on the class using the provided dict of values.

        Args:
        ----
                data dict[str, int | VariableEntry]: the dictionary of values to update on the class
        """
        # Use the dict provided to populate the properties in the class
        if data is not None:
            for name, value in data.items():
                private_name = '_' + name
                setattr(self, private_name, self.__wrap(value))

        # Prevent reload from disk
        self.__grid_properties_loaded = True

    def to_dict(self) -> dict[str, Optional[int] | VariableEntry]:
        self.load_grid_properties_if_not_loaded()

        original_dict = self.__dict__
        new_dict: dict[str, Optional[int] | VariableEntry] = {}
        for key in original_dict.keys():
            new_key = key
            if new_key[0] == '_':
                new_key = new_key.replace('_', '', 1)
            if new_key[0] == '_':
                new_key = new_key.replace('_', '', 1)
            new_dict[new_key] = original_dict[key]

        return new_dict

    def load_grid_properties_if_not_loaded(self) -> None:
        def move_next_value(next_line: str) -> tuple[str, str]:
            """Finds the next value and then strips out the value from the line.

            Args:
            ----
                next_line (str): the line to search through for the value

            Raises:
            ------
                ValueError: if no value is found within the line provided

            Returns:
            -------
                tuple[str, str]: the next value found in the line, the line with the value stripped out.
            """
            value = nfo.get_next_value(0, [next_line], next_line)
            if value is None:
                raise ValueError(f"No value found within the provided line: {next_line}")
            next_line = next_line.replace(value, "", 1)
            return value, next_line

        # If we've already loaded the grid properties, don't do so again.
        if self.__grid_properties_loaded:
            return

        if self.__grid_nexus_file is None or self.__grid_file_contents is None or self.__grid_file_nested is None:
            raise ValueError("Grid file not found, cannot load grid properties")

        file_as_list = self.__grid_file_contents
        for line in file_as_list:

            # Load in the basic properties
            properties_to_load = [
                PropertyToLoad('NETGRS', ['VALUE', 'CON'], self._netgrs),
                PropertyToLoad('POROSITY', ['VALUE', 'CON'], self._porosity),
                PropertyToLoad('SW', ['VALUE', 'CON'], self._sw),
                PropertyToLoad('KX', ['VALUE', 'MULT', 'CON'], self._kx),
                PropertyToLoad('PERMX', ['VALUE', 'MULT', 'CON'], self._kx),
                PropertyToLoad('PERMI', ['VALUE', 'MULT', 'CON'], self._kx),
                PropertyToLoad('KI', ['VALUE', 'MULT', 'CON'], self._kx),
                PropertyToLoad('KY', ['VALUE', 'MULT', 'CON'], self._ky),
                PropertyToLoad('PERMY', ['VALUE', 'MULT', 'CON'], self._ky),
                PropertyToLoad('PERMJ', ['VALUE', 'MULT', 'CON'], self._ky),
                PropertyToLoad('KJ', ['VALUE', 'MULT', 'CON'], self._ky),
                PropertyToLoad('KZ', ['VALUE', 'MULT', 'CON'], self._kz),
                PropertyToLoad('PERMZ', ['VALUE', 'MULT', 'CON'], self._kz),
                PropertyToLoad('PERMK', ['VALUE', 'MULT', 'CON'], self._kz),
                PropertyToLoad('KK', ['VALUE', 'MULT', 'CON'], self._kz),
            ]

            for token_property in properties_to_load:
                for modifier in token_property.modifiers:
                    StructuredGridOperations.load_token_value_if_present(
                        token_property.token, modifier, token_property.property, line, file_as_list, ['INCLUDE'])

            # Load in grid dimensions
            if nfo.check_token('NX', line):
                # Check that the format of the grid is NX followed by NY followed by NZ
                current_line = file_as_list[file_as_list.index(line)]
                remaining_line = current_line[current_line.index('NX') + 2:]
                if nfo.get_next_value(0, [remaining_line], remaining_line) != 'NY':
                    continue
                remaining_line = remaining_line[remaining_line.index('NY') + 2:]
                if nfo.get_next_value(0, [remaining_line], remaining_line) != 'NZ':
                    continue

                # Avoid loading in a comment
                if "!" in line and line.index("!") < line.index('NX'):
                    continue
                next_line = file_as_list[file_as_list.index(line) + 1]
                first_value, next_line = move_next_value(next_line)
                second_value, next_line = move_next_value(next_line)
                third_value, next_line = move_next_value(next_line)

                self._range_x = int(first_value)
                self._range_y = int(second_value)
                self._range_z = int(third_value)

        self.__grid_properties_loaded = True

    @classmethod
    def load_structured_grid_file(cls: type[NexusGrid], structured_grid_file: NexusFile,
                                  lazy_loading: bool = True) -> NexusGrid:
        """Loads in a structured grid file with all grid properties, and the array functions defined with 'FUNCTION'.
        Other grid modifiers are currently not supported.

        Args:
        ----
            structured_grid_file (NexusFile): the NexusFile representation of a structured grid file for converting \
                into a structured grid file class
        Raises:
            AttributeError: if no value is found for the structured grid file path
            ValueError: if when loading the grid no values can be found for the NX NY NZ line.
        """
        if structured_grid_file.location is None:
            raise ValueError(f"No file path given or found for structured grid file path. \
                Instead got {structured_grid_file.location}")

        loaded_structured_grid_file = cls(grid_nexus_file=structured_grid_file)

        if not lazy_loading:
            loaded_structured_grid_file.load_grid_properties_if_not_loaded()
            loaded_structured_grid_file.load_faults()
            loaded_structured_grid_file.load_array_functions()

        return loaded_structured_grid_file

    @staticmethod
    def update_structured_grid_file(grid_dict: dict[str, VariableEntry | int], model: NexusSimulator) -> None:
        """Save values passed from the front end to the structured grid file and update the class.

        Args:
        ----
            grid_dict (dict[str, Union[VariableEntry, int]]): dictionary containing grid properties to be replaced
            model (NexusSimulator): an instance of a NexusSimulator object
        Raises:
            ValueError: If no structured grid file is in the instance of the Simulator class
        """
        # Convert the dictionary back to a class, and update the properties on our class
        structured_grid = model.grid
        if structured_grid is None or model.model_files.structured_grid_file is None:
            raise ValueError("Model does not contain a structured grid")
        original_structured_grid_file = copy.deepcopy(structured_grid)

        # replace the structured grid with a new object with an updated dictionary
        structured_grid.update_properties_from_dict(grid_dict)

        # change it in the text file for nexus:
        grid_file_path = model.model_files.structured_grid_file.location
        if grid_file_path is None:
            raise ValueError("No path found for structured grid file path.")
        structured_grid_contents = nfo.load_file_as_list(grid_file_path)
        # Get the existing file as a list
        if structured_grid_contents is None:
            raise ValueError("No path found for structured grid file path. \
                Please provide a path to the structured grid")
        # Update each value in the file
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.netgrs,
                                               structured_grid.netgrs, 'NETGRS')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.porosity,
                                               structured_grid.porosity, 'POROSITY')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.sw,
                                               structured_grid.sw, 'SW')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.kx,
                                               structured_grid.kx, 'KX')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.ky,
                                               structured_grid.ky, 'KY')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.kz,
                                               structured_grid.kz, 'KZ')

        # Save the new file contents
        new_file_str = "".join(structured_grid_contents)
        with open(grid_file_path, "w") as text_file:
            text_file.write(new_file_str)

    def load_array_functions(self) -> None:
        if self.__grid_file_contents is None:
            raise ValueError("Cannot load array functions as grid file cannot not found")
        self.__array_functions_list = afo.collect_all_function_blocks(self.__grid_file_contents)
        self.__array_functions_df = afo.summarize_model_functions(self.__array_functions_list)
        self.__array_functions_loaded = True

    def get_array_functions_list(self) -> Optional[list[list[str]]]:
        """Returns the grid array functions as a list of function lines."""
        if not self.__array_functions_loaded:
            self.load_array_functions()
        return self.__array_functions_list

    def get_array_functions_df(self) -> Optional[pd.DataFrame]:
        """Returns the grid array functions as a dataframe."""
        if not self.__array_functions_loaded:
            self.load_array_functions()
        return self.__array_functions_df

    def load_faults(self) -> None:
        """Function to read faults in Nexus grid file defined using MULT and FNAME keywords."""
        file_content_as_list = self.__grid_file_contents
        if file_content_as_list is None:
            raise ValueError('Grid file contents have not been loaded')
        df = load_nexus_fault_mult_table_from_list(file_content_as_list)

        if not df.empty:
            # Ensure resulting dataframe has uppercase column names
            df.columns = [col.upper() for col in df.columns]

            # Check if any multfls have been used in grid file and update fault trans multipliers accordingly
            f_names = df['NAME'].unique()
            f_mults = [1.] * len(f_names)
            mult_dict = dict(zip(f_names, f_mults))
            for line in file_content_as_list:
                if nfo.check_token('MULTFL', line):
                    fname = str(nfo.get_expected_token_value(
                        'MULTFL', line, file_content_as_list,
                        custom_message=f'{line} does not have a fault name following MULTFL'))
                    if fname in df['NAME'].unique():
                        tmult = float(str(nfo.get_expected_token_value(
                            fname, line, file_content_as_list,
                            custom_message=f'MULTFL {fname} does not have a numerical tmult value')))
                        mult_dict[fname] *= tmult
            mult_df = pd.DataFrame.from_dict(mult_dict, orient='index').reset_index()
            mult_df.columns = ['NAME', 'TMULT']
            new_df = df.merge(mult_df, how='left', on='NAME', validate='many_to_one')
            new_df['MULT'] = new_df['MULT'] * new_df['TMULT']
            self.__faults_df = new_df.drop(['TMULT'], axis=1)
        self.__grid_faults_loaded = True

    def get_faults_df(self) -> Optional[pd.DataFrame]:
        """Returns the fault definition and transmissility multiplier information as a dataframe."""

        if not self.__grid_faults_loaded:
            self.load_faults()
        return self.__faults_df



"""Class that represents a single nexus target in the NexusSimulator."""
from __future__ import annotations
from dataclasses import dataclass
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Target import Target


@dataclass(kw_only=True, repr=False)
class NexusTarget(Target):
    """Class that represents a single nexus target in the NexusSimulator."""
    def __init__(self, properties_dict: dict[str, None | int | str | float]) -> None:
        # call the init of the DataObjectMixin
        super(Target, self).__init__({})
        for key, prop in properties_dict.items():
            self.__setattr__(key, prop)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords = {
            'NAME': ('name', str),
            'CTRL': ('control_quantity', str),
            'CTRLCOND': ('control_conditions', str),
            'CTRLCONS': ('control_connections', str),
            'CTRLMETHOD': ('control_method', str),
            'CALCMETHOD': ('calculation_method', str),
            'CALCCOND': ('calculation_conditions', str),
            'CALCCONS': ('calculation_connections', str),
            'VALUE': ('value', float),
            'ADDVALUE': ('add_value', float),
            'REGION': ('region', str),
            'PRIORITY': ('priority', int),
            'QMIN': ('minimum_rate', float),
            'QMIN_NOSHUT': ('minimum_rate_no_shut', float),
            'QGUIDE': ('guide_rate', str),
            'MAXDPDT': ('max_change_pressure', float),
            'RANKDT': ('rank_dt', float),
            'CTRLTYPE': ('control_type', str),
            'CALCTYPE': ('calculation_type', str)
            }
        return keywords

    def update(self, new_data: dict[str, None | int | str | float | UnitSystem], nones_overwrite: bool = False):
        """Updates attributes in the object based on the dictionary provided."""
        for k, v in new_data.items():
            if v is not None or nones_overwrite:
                setattr(self, k, v)



from abc import ABC
from dataclasses import field
from typing import Literal

from ResSimpy.OperationsMixin import NetworkOperationsMixIn
from ResSimpy.WellConnection import WellConnection


class WellConnections(NetworkOperationsMixIn, ABC):
    __well_connections: list[WellConnection] = field(default_factory=list)

    @property
    def _network_element_name(self) -> Literal['well_connections']:
        return 'well_connections'



"""Class for handling the mapping between unit systems and the units used for that unit system."""

from typing import Mapping

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.Units.Units import UnitDimension, SurfaceRatesLiquid, SurfaceRatesGas, ReservoirRates, MolarRates, \
    Pressure, Temperature, SaturationFraction, GasLiquidRatio, LiquidGasRatio, SurfaceVolumesGas, SurfaceVolumesLiquid


class ConstraintUnits(BaseUnitMapping):
    """Class for handling the mapping between unit systems and the units used for that unit system."""
    attribute_map: Mapping[str, UnitDimension] = {
        'max_surface_oil_rate': SurfaceRatesLiquid(),
        'max_surface_water_rate': SurfaceRatesLiquid(),
        'max_surface_gas_rate': SurfaceRatesGas(),
        'max_surface_liquid_rate': SurfaceRatesLiquid(),
        'max_reservoir_oil_rate': ReservoirRates(),
        'max_reservoir_water_rate': ReservoirRates(),
        'max_reservoir_gas_rate': ReservoirRates(),
        'max_reservoir_liquid_rate': ReservoirRates(),
        'max_hc_moles_rate': MolarRates(),
        'max_reverse_surface_oil_rate': SurfaceRatesLiquid(),
        'max_reverse_surface_water_rate': SurfaceRatesLiquid(),
        'max_reverse_surface_gas_rate': SurfaceRatesGas(),
        'max_reverse_surface_liquid_rate': SurfaceRatesLiquid(),
        'max_reverse_reservoir_oil_rate': ReservoirRates(),
        'max_reverse_reservoir_water_rate': ReservoirRates(),
        'max_reverse_reservoir_gas_rate': ReservoirRates(),
        'max_reverse_reservoir_liquid_rate': ReservoirRates(),
        'max_reverse_hc_moles_rate': MolarRates(),
        'min_pressure': Pressure(),
        'max_pressure': Pressure(),
        'min_temperature': Temperature(),
        'max_wag_water_pressure': Pressure(),
        'max_wag_gas_pressure': Pressure(),
        'bottom_hole_pressure': Pressure(),
        'tube_head_pressure': Pressure(),
        'min_surface_oil_rate': SurfaceRatesLiquid(),
        'min_surface_water_rate': SurfaceRatesLiquid(),
        'min_surface_gas_rate': SurfaceRatesGas(),
        'min_surface_liquid_rate': SurfaceRatesLiquid(),
        'min_reservoir_oil_rate': ReservoirRates(),
        'min_reservoir_water_rate': ReservoirRates(),
        'min_reservoir_gas_rate': ReservoirRates(),
        'min_reservoir_liquid_rate': ReservoirRates(),
        'min_hc_moles_rate': MolarRates(),
        'min_reverse_surface_oil_rate': SurfaceRatesLiquid(),
        'min_reverse_surface_water_rate': SurfaceRatesLiquid(),
        'min_reverse_surface_gas_rate': SurfaceRatesGas(),
        'min_reverse_surface_liquid_rate': SurfaceRatesLiquid(),
        'min_reverse_reservoir_oil_rate': ReservoirRates(),
        'min_reverse_reservoir_water_rate': ReservoirRates(),
        'min_reverse_reservoir_gas_rate': ReservoirRates(),
        'min_reverse_reservoir_liquid_rate': ReservoirRates(),
        'min_reverse_hc_moles_rate': MolarRates(),
        'max_watercut': SaturationFraction(),
        'max_watercut_plug': SaturationFraction(),
        'max_watercut_plugplus': SaturationFraction(),
        'max_watercut_perf': SaturationFraction(),
        'max_watercut_perfplus': SaturationFraction(),
        'max_wor': SaturationFraction(),
        'max_wor_plug': SaturationFraction(),
        'max_wor_plugplus': SaturationFraction(),
        'max_wor_perf': SaturationFraction(),
        'max_wor_perfplus': SaturationFraction(),
        'max_gor': GasLiquidRatio(),
        'max_gor_plug': GasLiquidRatio(),
        'max_gor_plugplus': GasLiquidRatio(),
        'max_gor_perf': GasLiquidRatio(),
        'max_gor_perfplus': GasLiquidRatio(),
        'max_lgr': LiquidGasRatio(),
        'max_lgr_plug': LiquidGasRatio(),
        'max_lgr_plugplus': LiquidGasRatio(),
        'max_lgr_perf': LiquidGasRatio(),
        'max_lgr_perfplus': LiquidGasRatio(),
        'max_cum_gas_prod': SurfaceVolumesGas(),
        'max_cum_oil_prod': SurfaceVolumesLiquid(),
        'max_cum_water_prod': SurfaceVolumesLiquid(),
        'max_cum_liquid_prod': SurfaceVolumesLiquid(),
        'qmult_oil_rate': SurfaceRatesLiquid(),
        'qmult_water_rate': SurfaceRatesLiquid(),
        'qmult_gas_rate': SurfaceRatesGas(),
    }

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    @property
    def max_surface_oil_rate(self) -> str:
        """Returns the unit for max_surface_oil_rate."""
        return self.get_unit_for_attribute('max_surface_oil_rate')

    @property
    def max_surface_water_rate(self) -> str:
        """Returns the unit for max_surface_water_rate."""
        return self.get_unit_for_attribute('max_surface_water_rate')

    @property
    def max_surface_gas_rate(self) -> str:
        """Returns the unit for max_surface_gas_rate."""
        return self.get_unit_for_attribute('max_surface_gas_rate')

    @property
    def max_surface_liquid_rate(self) -> str:
        """Returns the unit for max_surface_liquid_rate."""
        return self.get_unit_for_attribute('max_surface_liquid_rate')

    @property
    def max_reservoir_oil_rate(self) -> str:
        """Returns the unit for max_reservoir_oil_rate."""
        return self.get_unit_for_attribute('max_reservoir_oil_rate')

    @property
    def max_reservoir_water_rate(self) -> str:
        """Returns the unit for max_reservoir_water_rate."""
        return self.get_unit_for_attribute('max_reservoir_water_rate')

    @property
    def max_reservoir_gas_rate(self) -> str:
        """Returns the unit for max_reservoir_gas_rate."""
        return self.get_unit_for_attribute('max_reservoir_gas_rate')

    @property
    def max_reservoir_liquid_rate(self) -> str:
        """Returns the unit for max_reservoir_liquid_rate."""
        return self.get_unit_for_attribute('max_reservoir_liquid_rate')

    @property
    def max_hc_moles_rate(self) -> str:
        """Returns the unit for max_hc_moles_rate."""
        return self.get_unit_for_attribute('max_hc_moles_rate')

    @property
    def max_reverse_surface_oil_rate(self) -> str:
        """Returns the unit for max_reverse_surface_oil_rate."""
        return self.get_unit_for_attribute('max_reverse_surface_oil_rate')

    @property
    def max_reverse_surface_water_rate(self) -> str:
        """Returns the unit for max_reverse_surface_water_rate."""
        return self.get_unit_for_attribute('max_reverse_surface_water_rate')

    @property
    def max_reverse_surface_gas_rate(self) -> str:
        """Returns the unit for max_reverse_surface_gas_rate."""
        return self.get_unit_for_attribute('max_reverse_surface_gas_rate')

    @property
    def max_reverse_surface_liquid_rate(self) -> str:
        """Returns the unit for max_reverse_surface_liquid_rate."""
        return self.get_unit_for_attribute('max_reverse_surface_liquid_rate')

    @property
    def max_reverse_reservoir_oil_rate(self) -> str:
        """Returns the unit for max_reverse_reservoir_oil_rate."""
        return self.get_unit_for_attribute('max_reverse_reservoir_oil_rate')

    @property
    def max_reverse_reservoir_water_rate(self) -> str:
        """Returns the unit for max_reverse_reservoir_water_rate."""
        return self.get_unit_for_attribute('max_reverse_reservoir_water_rate')

    @property
    def max_reverse_reservoir_gas_rate(self) -> str:
        """Returns the unit for max_reverse_reservoir_gas_rate."""
        return self.get_unit_for_attribute('max_reverse_reservoir_gas_rate')

    @property
    def max_reverse_reservoir_liquid_rate(self) -> str:
        """Returns the unit for max_reverse_reservoir_liquid_rate."""
        return self.get_unit_for_attribute('max_reverse_reservoir_liquid_rate')

    @property
    def max_reverse_hc_moles_rate(self) -> str:
        """Returns the unit for max_reverse_hc_moles_rate."""
        return self.get_unit_for_attribute('max_reverse_hc_moles_rate')

    @property
    def min_pressure(self) -> str:
        """Returns the unit for min_pressure."""
        return self.get_unit_for_attribute('min_pressure')

    @property
    def max_pressure(self) -> str:
        """Returns the unit for max_pressure."""
        return self.get_unit_for_attribute('max_pressure')

    @property
    def min_temperature(self) -> str:
        """Returns the unit for min_temperature."""
        return self.get_unit_for_attribute('min_temperature')

    @property
    def max_wag_water_pressure(self) -> str:
        """Returns the unit for max_wag_water_pressure."""
        return self.get_unit_for_attribute('max_wag_water_pressure')

    @property
    def max_wag_gas_pressure(self) -> str:
        """Returns the unit for max_wag_gas_pressure."""
        return self.get_unit_for_attribute('max_wag_gas_pressure')

    @property
    def bottom_hole_pressure(self) -> str:
        """Returns the unit for bottom_hole_pressure."""
        return self.get_unit_for_attribute('bottom_hole_pressure')

    @property
    def tube_head_pressure(self) -> str:
        """Returns the unit for tube_head_pressure."""
        return self.get_unit_for_attribute('tube_head_pressure')

    @property
    def min_surface_oil_rate(self) -> str:
        """Returns the unit for min_surface_oil_rate."""
        return self.get_unit_for_attribute('min_surface_oil_rate')

    @property
    def min_surface_water_rate(self) -> str:
        """Returns the unit for min_surface_water_rate."""
        return self.get_unit_for_attribute('min_surface_water_rate')

    @property
    def min_surface_gas_rate(self) -> str:
        """Returns the unit for min_surface_gas_rate."""
        return self.get_unit_for_attribute('min_surface_gas_rate')

    @property
    def min_surface_liquid_rate(self) -> str:
        """Returns the unit for min_surface_liquid_rate."""
        return self.get_unit_for_attribute('min_surface_liquid_rate')

    @property
    def min_reservoir_oil_rate(self) -> str:
        """Returns the unit for min_reservoir_oil_rate."""
        return self.get_unit_for_attribute('min_reservoir_oil_rate')

    @property
    def min_reservoir_water_rate(self) -> str:
        """Returns the unit for min_reservoir_water_rate."""
        return self.get_unit_for_attribute('min_reservoir_water_rate')

    @property
    def min_reservoir_gas_rate(self) -> str:
        """Returns the unit for min_reservoir_gas_rate."""
        return self.get_unit_for_attribute('min_reservoir_gas_rate')

    @property
    def min_reservoir_liquid_rate(self) -> str:
        """Returns the unit for min_reservoir_liquid_rate."""
        return self.get_unit_for_attribute('min_reservoir_liquid_rate')

    @property
    def min_hc_moles_rate(self) -> str:
        """Returns the unit for min_hc_moles_rate."""
        return self.get_unit_for_attribute('min_hc_moles_rate')

    @property
    def min_reverse_surface_oil_rate(self) -> str:
        """Returns the unit for min_reverse_surface_oil_rate."""
        return self.get_unit_for_attribute('min_reverse_surface_oil_rate')

    @property
    def min_reverse_surface_water_rate(self) -> str:
        """Returns the unit for min_reverse_surface_water_rate."""
        return self.get_unit_for_attribute('min_reverse_surface_water_rate')

    @property
    def min_reverse_surface_gas_rate(self) -> str:
        """Returns the unit for min_reverse_surface_gas_rate."""
        return self.get_unit_for_attribute('min_reverse_surface_gas_rate')

    @property
    def min_reverse_surface_liquid_rate(self) -> str:
        """Returns the unit for min_reverse_surface_liquid_rate."""
        return self.get_unit_for_attribute('min_reverse_surface_liquid_rate')

    @property
    def min_reverse_reservoir_oil_rate(self) -> str:
        """Returns the unit for min_reverse_reservoir_oil_rate."""
        return self.get_unit_for_attribute('min_reverse_reservoir_oil_rate')

    @property
    def min_reverse_reservoir_water_rate(self) -> str:
        """Returns the unit for min_reverse_reservoir_water_rate."""
        return self.get_unit_for_attribute('min_reverse_reservoir_water_rate')

    @property
    def min_reverse_reservoir_gas_rate(self) -> str:
        """Returns the unit for min_reverse_reservoir_gas_rate."""
        return self.get_unit_for_attribute('min_reverse_reservoir_gas_rate')

    @property
    def min_reverse_reservoir_liquid_rate(self) -> str:
        """Returns the unit for min_reverse_reservoir_liquid_rate."""
        return self.get_unit_for_attribute('min_reverse_reservoir_liquid_rate')

    @property
    def min_reverse_hc_moles_rate(self) -> str:
        """Returns the unit for min_reverse_hc_moles_rate."""
        return self.get_unit_for_attribute('min_reverse_hc_moles_rate')

    @property
    def max_watercut(self) -> str:
        """Returns the unit for max_watercut."""
        return self.get_unit_for_attribute('max_watercut')

    @property
    def max_watercut_plug(self) -> str:
        """Returns the unit for max_watercut_plug."""
        return self.get_unit_for_attribute('max_watercut_plug')

    @property
    def max_watercut_plugplus(self) -> str:
        """Returns the unit for max_watercut_plugplus."""
        return self.get_unit_for_attribute('max_watercut_plugplus')

    @property
    def max_watercut_perf(self) -> str:
        """Returns the unit for max_watercut_perf."""
        return self.get_unit_for_attribute('max_watercut_perf')

    @property
    def max_watercut_perfplus(self) -> str:
        """Returns the unit for max_watercut_perfplus."""
        return self.get_unit_for_attribute('max_watercut_perfplus')

    @property
    def max_wor(self) -> str:
        """Returns the unit for max_wor."""
        return self.get_unit_for_attribute('max_wor')

    @property
    def max_wor_plug(self) -> str:
        """Returns the unit for max_wor_plug."""
        return self.get_unit_for_attribute('max_wor_plug')

    @property
    def max_wor_plugplus(self) -> str:
        """Returns the unit for max_wor_plugplus."""
        return self.get_unit_for_attribute('max_wor_plugplus')

    @property
    def max_wor_perf(self) -> str:
        """Returns the unit for max_wor_perf."""
        return self.get_unit_for_attribute('max_wor_perf')

    @property
    def max_wor_perfplus(self) -> str:
        """Returns the unit for max_wor_perfplus."""
        return self.get_unit_for_attribute('max_wor_perfplus')

    @property
    def max_gor(self) -> str:
        """Returns the unit for max_gor."""
        return self.get_unit_for_attribute('max_gor')

    @property
    def max_gor_plug(self) -> str:
        """Returns the unit for max_gor_plug."""
        return self.get_unit_for_attribute('max_gor_plug')

    @property
    def max_gor_plugplus(self) -> str:
        """Returns the unit for max_gor_plugplus."""
        return self.get_unit_for_attribute('max_gor_plugplus')

    @property
    def max_gor_perf(self) -> str:
        """Returns the unit for max_gor_perf."""
        return self.get_unit_for_attribute('max_gor_perf')

    @property
    def max_gor_perfplus(self) -> str:
        """Returns the unit for max_gor_perfplus."""
        return self.get_unit_for_attribute('max_gor_perfplus')

    @property
    def max_lgr(self) -> str:
        """Returns the unit for max_lgr."""
        return self.get_unit_for_attribute('max_lgr')

    @property
    def max_lgr_plug(self) -> str:
        """Returns the unit for max_lgr_plug."""
        return self.get_unit_for_attribute('max_lgr_plug')

    @property
    def max_lgr_plugplus(self) -> str:
        """Returns the unit for max_lgr_plugplus."""
        return self.get_unit_for_attribute('max_lgr_plugplus')

    @property
    def max_lgr_perf(self) -> str:
        """Returns the unit for max_lgr_perf."""
        return self.get_unit_for_attribute('max_lgr_perf')

    @property
    def max_lgr_perfplus(self) -> str:
        """Returns the unit for max_lgr_perfplus."""
        return self.get_unit_for_attribute('max_lgr_perfplus')

    @property
    def max_cum_gas_prod(self) -> str:
        """Returns the unit for max_cum_gas_prod."""
        return self.get_unit_for_attribute('max_cum_gas_prod')

    @property
    def max_cum_oil_prod(self) -> str:
        """Returns the unit for max_cum_oil_prod."""
        return self.get_unit_for_attribute('max_cum_oil_prod')

    @property
    def max_cum_water_prod(self) -> str:
        """Returns the unit for max_cum_water_prod."""
        return self.get_unit_for_attribute('max_cum_water_prod')

    @property
    def max_cum_liquid_prod(self) -> str:
        """Returns the unit for max_cum_liquid_prod."""
        return self.get_unit_for_attribute('max_cum_liquid_prod')

    @property
    def qmult_oil_rate(self) -> str:
        """Returns the unit for qmult_oil_rate."""
        return self.get_unit_for_attribute('qmult_oil_rate')

    @property
    def qmult_water_rate(self) -> str:
        """Returns the unit for qmult_water_rate."""
        return self.get_unit_for_attribute('qmult_water_rate')

    @property
    def qmult_gas_rate(self) -> str:
        """Returns the unit for qmult_gas_rate."""
        return self.get_unit_for_attribute('qmult_gas_rate')



"""Class to hold Nexus relative permeability and capillary pressure properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.relpm_keywords import RELPM_TABLE_KEYWORDS, RELPM_KEYWORDS_VALUE_FLOAT
from ResSimpy.Nexus.NexusKeywords.relpm_keywords import RELPM_SINGLE_KEYWORDS, RELPM_HYSTERESIS_PRIMARY_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.relpm_keywords import RELPM_KEYWORDS, RELPM_NONDARCY_KEYWORDS, RELPM_NONDARCY_PARAMS
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import RelPermUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union, get_empty_hysteresis_dict
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusRelPermMethod(DynamicProperty):
    """Class to hold Nexus relative permeability and capillary pressure properties.

    Attributes:
        file (NexusFile): Nexus relperm file object
        input_number (int): RELPM method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific PVT method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                     dict[str, Union[float, pd.DataFrame]]]] = field(default_factory=get_empty_dict_union)
    hysteresis_params: dict[str, Union[str, float, dict[str, Union[str, float, dict[str, Union[str, float]]]]]] \
        = field(default_factory=get_empty_hysteresis_dict)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None,
                 hysteresis_params: Optional[dict[str, Union[str, float, dict[str, Union[str, float,
                                             dict[str, Union[str, float]]]]]]] = None) -> None:
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        if hysteresis_params is not None:
            self.hysteresis_params = hysteresis_params
        else:
            self.hysteresis_params = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'PCWO': ('water_oil_capillary_pressure', float),
            'PCGO': ('gas_oil_capillary_pressure', float),
            'PCGW': ('gas_water_capillary_pressure', float),
            'TENTHR': ('interfacial_tension_threshold_for_relperm_adjustment', float),
            'TENI': ('reference_interfacial_tension_for_capillary_pressure_adjustment', float),
        }
        return keywords

    @property
    def units(self) -> RelPermUnits:
        """Returns the attribute to unit map for the relperm method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return RelPermUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with relative permeability and capillary pressure data, in Nexus file format."""
        printable_str = ''
        # Handle non-hysteresis relperm and capillary pressure parameters
        relperm_dict = self.properties
        stone1_options = {'STONE1': ['SOMOPT1', 'SOMOPT2', 'SOMOPT3', 'ST1EXP', 'ST1EPS', 'ST1NU'],
                          'STONE1_WAT': ['SWMOPT1', 'SWMOPT2']}
        for key, value in relperm_dict.items():
            if isinstance(value, pd.DataFrame):
                table_text = f'{key}'
                if key == 'WOTABLE':
                    if 'LOW_SAL' in relperm_dict.keys():
                        table_text += " LOW_SAL"
                printable_str += f'{table_text}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n\n'
            elif key in ['STONE1', 'STONE1_WAT']:
                printable_str += f'{key}'
                for stone1_key in stone1_options[key]:
                    if stone1_key in relperm_dict.keys():
                        if relperm_dict[stone1_key] == '':
                            printable_str += f' {stone1_key}'
                        else:
                            printable_str += f' {stone1_key} {relperm_dict[stone1_key]}'
                printable_str += '\n'
            elif key == 'JFUNC':
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
                for jfunc_key in ['PERMBASIS', 'PORBASIS']:
                    if jfunc_key in relperm_dict.keys():
                        printable_str += f'       {jfunc_key} {relperm_dict[jfunc_key]}\n'
            elif key in ['SCALING', 'SCALING_PC', 'IFT', 'NEARCRIT', 'DERIVATIVES']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
            elif isinstance(value, dict):
                printable_str += f'{key}\n'
                for subkey, subvalue in value.items():
                    printable_str += f'    {subkey} {subvalue}\n'
                if key in RELPM_NONDARCY_KEYWORDS:
                    printable_str += 'END'+key+'\n'
            elif key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key not in ['DESC', 'LOW_SAL', 'STONE1', 'SOMOPT1', 'SOMOPT2', 'SOMOPT3', 'ST1EXP', 'ST1EPS', 'ST1NU',
                             'STONE1_WAT', 'SWMOPT1', 'SWMOPT2', 'PERMBASIS', 'PORBASIS']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'

        # Handle non-hysteresis relperm and capillary pressure parameters
        hyst_dict = self.hysteresis_params
        key_indx = 0
        for key, value in hyst_dict.items():
            if key_indx == 0:
                printable_str += 'HYSTERESIS'
            else:
                printable_str += '          '
            if value == '':
                printable_str += f' {key}\n'
            elif value == 'USER':
                printable_str += f' {key} USER\n'
            elif isinstance(value, dict):
                printable_str += f' {key}'
                for subkey, subvalue in value.items():
                    if isinstance(subvalue, dict):
                        printable_str += f' {subkey}'
                        for subsubkey, subsubvalue in subvalue.items():
                            if subsubvalue == '':
                                printable_str += f' {subsubkey}'
                            else:
                                printable_str += f' {subsubkey} {subsubvalue}'
                    else:
                        if subvalue == '':
                            printable_str += f' {subkey}'
                        else:
                            printable_str += f' {subkey} {subvalue}'
                printable_str += '\n'
            else:
                printable_str += f' {key} {value}\n'
            key_indx += 1
        return printable_str

    def __populate_optional_str_keywords(self, keyword: str, keyword_value_options: list[str], single_line: str,
                                         line_list: list[str]) -> None:
        """Utility function to populate rel perm keywords that have optional string values, e.g., IFT, JFUNC, etc.

        Args:
        ----
            keyword (str): primary keyword, e.g., IFT or JFUNC
            keyword_value_options (list[str]): primary keyword optional values, e.g., [METHOD1, METHOD2] for IFT
            single_line (str): single line as read from input RELPM file
            line_list (list[str]): list of strings that comprise input RELPM file
        """
        if nfo.check_token(keyword, single_line):
            key_val = nfo.get_token_value(keyword, single_line, line_list)
            if key_val in keyword_value_options:
                self.properties[keyword] = key_val
            else:
                self.properties[keyword] = ''

    def read_properties(self) -> None:
        """Read Nexus rel perm file contents and populate the NexusRelPermMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through relperm file contents
        # Dictionary to record start and ending indices for tables
        relpm_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in RELPM_TABLE_KEYWORDS:
            table_being_read[table_name] = False
        # Flag to track hysteresis parameters section
        hysteresis_being_read: bool = False
        hysteresis_section_indices: list[int] = []
        # Dictionary to record start and ending indices of nondarcy section
        nondarcy_indices: dict[str, list[int]] = {}
        nondarcy_being_read: bool = False
        # Dictionary and flags for handle tabular reconstruction
        recon_dict: dict[str, Union[float, pd.DataFrame]] = {}
        found_reconstruct = False

        line_indx = 0
        for line in file_as_list:

            # Find standalone relperm keywords
            if [i for i in line.split() if i in RELPM_SINGLE_KEYWORDS]:
                for word in RELPM_SINGLE_KEYWORDS:
                    if nfo.check_token(word, line):
                        self.properties[word] = ''
            # Find relperm key-value pairs, such as SOMOPT2 0.05
            if [i for i in line.split() if i in RELPM_KEYWORDS_VALUE_FLOAT]:
                for key in RELPM_KEYWORDS_VALUE_FLOAT:
                    if nfo.check_token(key, line):
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            # Handle certain specific relperm options
            optional_keyword_dict = {
                'JFUNC': ['KX', 'KY', 'KXKY'],
                'SCALING': ['NONE', 'TWOPOINT', 'THREEPOINT'],
                'SCALING_PC': ['NONE', 'TWOPOINT', 'THREEPOINT'],
                'IFT': ['METHOD1', 'METHOD2'],
                'NEARCRIT': ['HCSCALE'],
                'DERIVATIVES': ['ANALYTICAL', 'NUMERICAL']
                }
            for opt_key, opt_vals in optional_keyword_dict.items():
                self.__populate_optional_str_keywords(opt_key, opt_vals, line, file_as_list)
            # Handle tabular reconstruction
            if nfo.check_token('RECONSTRUCT', line):
                found_reconstruct = True
            for recon_key in ['NSGDIM', 'NSWDIM']:
                if nfo.check_token(recon_key, line) and found_reconstruct:
                    recon_dict[recon_key] = int(nfo.get_expected_token_value(recon_key, line, file_as_list))
            if found_reconstruct:
                self.properties['RECONSTRUCT'] = recon_dict

            # Find beginning and ending indices of hysteresis section
            if nfo.check_token('HYSTERESIS', line):
                hysteresis_being_read = True
                hysteresis_section_indices = [line_indx, len(file_as_list)]
            if len(line.split()) > 0 and hysteresis_being_read:
                possible_token = line.split()[0]
                if nfo.check_token(possible_token, line) and possible_token not in \
                        ["HYSTERESIS", *RELPM_HYSTERESIS_PRIMARY_KEYWORDS]:
                    hysteresis_being_read = False
                    hysteresis_section_indices[1] = line_indx

            # Find beginning and ending indices of nondarcy section
            for nondarcy_key in RELPM_NONDARCY_KEYWORDS:
                if nfo.check_token(nondarcy_key, line):
                    nondarcy_being_read = True
                    nondarcy_indices[nondarcy_key] = [line_indx + 1, len(file_as_list)]
                if nondarcy_being_read and nfo.check_token('END' + nondarcy_key, line):
                    nondarcy_being_read = False
                    nondarcy_indices[nondarcy_key][1] = line_indx

            # Find ending index of relperm tables
            # if start_reading_table:
            for table_keyword in RELPM_TABLE_KEYWORDS:
                if table_being_read[table_keyword]:
                    for potential_endkeyword in RELPM_KEYWORDS:
                        if nfo.check_token(potential_endkeyword, line):
                            relpm_table_indices[table_keyword][1] = line_indx
                            table_being_read[table_keyword] = False
                            break
            # Find the starting index of relperm tables
            if [i for i in line.split() if i in RELPM_TABLE_KEYWORDS]:
                for table_keyword in RELPM_TABLE_KEYWORDS:
                    if nfo.check_token(table_keyword, line):
                        relpm_table_indices[table_keyword] = [line_indx + 1, len(file_as_list)]
                        table_being_read[table_keyword] = True

            line_indx += 1

        # Read in tables
        for key in relpm_table_indices.keys():
            self.properties[key] = nfo.read_table_to_df(file_as_list[
                                                        relpm_table_indices[key][0]:relpm_table_indices[key][1]])

        # Read nondarcy parameters from nondarcy section, if present:
        for key in nondarcy_indices.keys():
            nondarcy_dict: dict[str, Union[float, pd.DataFrame]] = {}
            for line in file_as_list[nondarcy_indices[key][0]:nondarcy_indices[key][1]]:
                for param in RELPM_NONDARCY_PARAMS:
                    if nfo.check_token(param, line):
                        nondarcy_dict[param] = float(nfo.get_expected_token_value(param, line, file_as_list))
            self.properties[key] = nondarcy_dict

        # Read hysteresis section, if present
        if len(hysteresis_section_indices) > 0:
            for line in file_as_list[hysteresis_section_indices[0]:hysteresis_section_indices[1]]:
                if [i for i in line.split() if i in ['NONE', 'NOCHK_HYS']]:
                    for hyst_keyword in ['NONE', 'NOCHK_HYS']:
                        if nfo.check_token(hyst_keyword, line):
                            self.hysteresis_params[hyst_keyword] = ''
                if nfo.check_token('KRW', line) and nfo.get_token_value('KRW', line, file_as_list) == 'USER':
                    self.hysteresis_params['KRW'] = 'USER'
                if [i for i in line.split() if i in ['KRG', 'KROW']]:
                    for hyst_keyword in ['KRG', 'KROW']:
                        if nfo.check_token(hyst_keyword, line):
                            hyst_keyword_primary_key = nfo.get_token_value(hyst_keyword, line, file_as_list)
                            if hyst_keyword_primary_key == 'USER':
                                self.hysteresis_params[hyst_keyword] = 'USER'
                            elif hyst_keyword_primary_key in ['LINEAR', 'SCALED', 'CARLSON', 'KILLOUGH']:
                                hyst_dict: dict[str, Union[str, float, dict[str, Union[str, float]]]] = {}
                                hyst_subdict: dict[str, Union[str, float]] = {}
                                if nfo.check_token('MAXTRAP', line):
                                    hyst_subdict['MAXTRAP'] = float(nfo.get_expected_token_value('MAXTRAP',
                                                                                                 line, file_as_list))
                                    hyst_dict[hyst_keyword_primary_key] = hyst_subdict
                                if nfo.check_token('EXP', line):
                                    hyst_subdict['EXP'] = float(nfo.get_expected_token_value('EXP',
                                                                                             line, file_as_list))
                                    hyst_dict[hyst_keyword_primary_key] = hyst_subdict
                                if nfo.check_token('NOMOD', line):
                                    hyst_subdict['NOMOD'] = ''
                                    hyst_dict[hyst_keyword_primary_key] = hyst_subdict
                                self.hysteresis_params[hyst_keyword] = hyst_dict
                if [i for i in line.split() if i in ['PCWO', 'PCGO', 'WAG']]:
                    for hyst_keyword in ['PCWO', 'PCGO', 'WAG']:
                        if nfo.check_token(hyst_keyword, line):
                            hyst_pc_dict: dict[str, Union[str, float, dict[str, Union[str, float]]]] = {}
                            # Search for options with key-value pairs
                            for hyst_keyword_primary_key in ['MAXSW', 'MINSG', 'ETA', 'LAND', 'ALPHA', 'AFAC']:
                                if nfo.check_token(hyst_keyword_primary_key, line):
                                    hyst_pc_dict[hyst_keyword_primary_key] = \
                                        float(nfo.get_expected_token_value(hyst_keyword_primary_key,
                                                                           line, file_as_list))
                            # Search for single keyword options
                            for hyst_keyword_primary_key in ['TRAPSCALE', 'NOWATHYS', 'NOGASHYS', 'NOOILHYS']:
                                if nfo.check_token(hyst_keyword_primary_key, line):
                                    hyst_pc_dict[hyst_keyword_primary_key] = ''
                            self.hysteresis_params[hyst_keyword] = hyst_pc_dict
                if [i for i in line.split() if i in ['TOLREV', 'TOLHYS']]:
                    for hyst_keyword in ['TOLREV', 'TOLHYS']:
                        if nfo.check_token(hyst_keyword, line):
                            self.hysteresis_params[hyst_keyword] = float(
                                nfo.get_expected_token_value(hyst_keyword, line, file_as_list))



from __future__ import annotations
from dataclasses import dataclass
from abc import ABC
from typing import Optional


@dataclass
class RelPermEndPoint(ABC):
    __swl: Optional[float] = None
    __swr: Optional[float] = None
    __swu: Optional[float] = None
    __sgl: Optional[float] = None
    __sgr: Optional[float] = None
    __sgu: Optional[float] = None
    __swro: Optional[float] = None
    __sgro: Optional[float] = None
    __sgrw: Optional[float] = None
    __krw_swro: Optional[float] = None
    __krw_swu: Optional[float] = None
    __krg_sgro: Optional[float] = None
    __krg_sgu: Optional[float] = None
    __kro_swl: Optional[float] = None
    __kro_swr: Optional[float] = None
    __kro_sgl: Optional[float] = None
    __kro_sgr: Optional[float] = None
    __krw_sgl: Optional[float] = None
    __krw_sgr: Optional[float] = None
    __krg_sgrw: Optional[float] = None
    __sgtr: Optional[float] = None
    __sotr: Optional[float] = None

    def __init__(self, swl: Optional[float] = None, swr: Optional[float] = None, swu: Optional[float] = None,
                 sgl: Optional[float] = None, sgr: Optional[float] = None, sgu: Optional[float] = None,
                 swro: Optional[float] = None, sgro: Optional[float] = None, sgrw: Optional[float] = None,
                 krw_swro: Optional[float] = None, krw_swu: Optional[float] = None, krg_sgro: Optional[float] = None,
                 krg_sgu: Optional[float] = None, kro_swl: Optional[float] = None, kro_swr: Optional[float] = None,
                 kro_sgl: Optional[float] = None, kro_sgr: Optional[float] = None, krw_sgl: Optional[float] = None,
                 krw_sgr: Optional[float] = None, krg_sgrw: Optional[float] = None, sgtr: Optional[float] = None,
                 sotr: Optional[float] = None) -> None:
        self.__swl = swl
        self.__swr = swr
        self.__swu = swu
        self.__sgl = sgl
        self.__sgr = sgr
        self.__sgu = sgu
        self.__swro = swro
        self.__sgro = sgro
        self.__sgrw = sgrw
        self.__krw_swro = krw_swro
        self.__krw_swu = krw_swu
        self.__krg_sgro = krg_sgro
        self.__krg_sgu = krg_sgu
        self.__kro_swl = kro_swl
        self.__kro_swr = kro_swr
        self.__kro_sgl = kro_sgl
        self.__kro_sgr = kro_sgr
        self.__krw_sgl = krw_sgl
        self.__krw_sgr = krw_sgr
        self.__krg_sgrw = krg_sgrw
        self.__sgtr = sgtr
        self.__sotr = sotr

    def to_dict(self) -> dict[str, Optional[float]]:
        attribute_dict: dict[str, Optional[float]] = {
            'swl': self.__swl,
            'swr': self.__swr,
            'swu': self.__swu,
            'sgl': self.__sgl,
            'sgr': self.__sgr,
            'sgu': self.__sgu,
            'swro': self.__swro,
            'sgro': self.__sgro,
            'sgrw': self.__sgrw,
            'krw_swro': self.__krw_swro,
            'krw_swu': self.__krw_swu,
            'krg_sgro': self.__krg_sgro,
            'krg_sgu': self.__krg_sgu,
            'kro_swl': self.__kro_swl,
            'kro_swr': self.__kro_swr,
            'kro_sgl': self.__kro_sgl,
            'kro_sgr': self.__kro_sgr,
            'krw_sgl': self.__krw_sgl,
            'krw_sgr': self.__krw_sgr,
            'krg_sgrw': self.__krg_sgrw,
            'sgtr': self.__sgtr,
            'sotr': self.__sotr,
        }
        return attribute_dict



"""Class storing Nexus Equilibration method properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Union, Optional
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Enums.UnitsEnum import SUnits, TemperatureUnits, UnitSystem
from ResSimpy.Nexus.NexusKeywords.equil_keywords import EQUIL_INTSAT_KEYWORDS, EQUIL_KEYWORDS_VALUE_FLOAT
from ResSimpy.Nexus.NexusKeywords.equil_keywords import EQUIL_TABLE_KEYWORDS, EQUIL_SINGLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.equil_keywords import EQUIL_COMPOSITION_OPTIONS
from ResSimpy.Nexus.NexusKeywords.equil_keywords import EQUIL_KEYWORDS
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import EquilUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusEquilMethod(DynamicProperty):
    """Class to hold Nexus Equil properties.

    Attributes:
        file (NexusFile): Nexus equilibration file object
        input_number (int): Equilibration method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific equilibration method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str],
                                pd.DataFrame, dict[str, Union[float, pd.DataFrame]]]] \
        = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'PINIT': ('initial_pressure', float),
            'DINIT': ('datum_depth', float),
            'DEPTH': ('depth', float),
            'X': ('x', float),
            'Y': ('y', float),
            'TINIT': ('initial_temperature', float),
            'TEMP': ('temperature', float),
            'GOC': ('gas_oil_contact_depth', float),
            'WOC': ('water_oil_contact_depth', float),
            'GOC_PALEO': ('gas_oil_contact_depth', float),
            'WOC_PALEO': ('water_oil_contact_depth', float),
            'GWC': ('gas_water_contact_depth', float),
            'PCGOC': ('gas_oil_capillary_pressure_at_gas_oil_contact', float),
            'PCWOC': ('water_oil_capillary_pressure_at_water_oil_contact', float),
            'PCGWC': ('gas_water_capillary_pressure_at_gas_water_contact', float),
            'PSAT': ('saturation_pressure', float),
            'API': ('oil_api_gravity', float),
        }
        return keywords

    @property
    def units(self) -> EquilUnits:
        """Returns the attribute to unit map for the Equil method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return EquilUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with equilibration data in Nexus file format."""
        printable_str = ''
        equil_dict = self.properties
        for key, value in equil_dict.items():
            if isinstance(value, pd.DataFrame):
                table_text = f'{key}'
                if key == 'COMPOSITION':
                    if 'X' in equil_dict.keys():
                        table_text += f" X {equil_dict['X']}"
                    if 'Y' in equil_dict.keys():
                        table_text += f" Y {equil_dict['Y']}"
                printable_str += f'{table_text}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif value == '':
                printable_str += f'{key}\n'
            elif key in ['INTSAT', 'VAITS']:
                if value == 'MOBILE':
                    printable_str += f'{key}\n    MOBILE\n'
                    for mobkey in ['SORWMN', 'SORGMN', 'SGCMN']:
                        if mobkey in equil_dict.keys():
                            printable_str += f'        {mobkey} {equil_dict[mobkey]}\n'
                for vaitkey in ['VAITS_TOLSG', 'VAITS_TOLSW']:
                    if vaitkey in equil_dict.keys():
                        printable_str += f'    {vaitkey} {equil_dict[vaitkey]}\n'
            elif key == 'OVERREAD':
                if isinstance(value, list):
                    printable_str += f"OVERREAD {' '.join(value)}\n"
            elif key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif key not in ['DESC', 'SORWMN', 'SORGMN', 'SGCMN', 'VAITS_TOLSG', 'VAITS_TOLSW',
                             'OVERREAD', 'INTSAT', 'VAITS', 'MOBILE', 'X', 'Y']:
                printable_str += f'{key} {value}\n'
        printable_str += '\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus equilibration file contents and populate NexusEquilMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize properties
        equil_table_indices: dict[str, list[int]] = {}
        # Flag to tell when to start reading a table
        start_reading_table: bool = False
        # Indicator of which of equilibration tables are being read
        table_being_read = ''

        line_indx = 0
        for line in file_as_list:

            # Find EQUIL key-value pairs, such as PINIT 3000, WOC 7000 or OVERREAD SW (list, multiple OVERREADs)
            if [i for i in line.split() if i in EQUIL_KEYWORDS_VALUE_FLOAT]:
                for key in EQUIL_KEYWORDS_VALUE_FLOAT:
                    if nfo.check_token(key, line) and 'DEPTH' not in line.split():  # Ensure not a table header
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            if nfo.check_token('AUTOGOC_COMP', line):
                self.properties['AUTOGOC_COMP'] = nfo.get_expected_token_value('AUTOGOC_COMP', line, file_as_list)
            if nfo.check_token('OVERREAD', line):
                overread_vals = line.split('!')[0].split('OVERREAD')[1].split()
                if 'OVERREAD' in self.properties.keys() and isinstance(self.properties['OVERREAD'], list):
                    self.properties['OVERREAD'] += overread_vals
                else:
                    self.properties['OVERREAD'] = overread_vals
            if nfo.check_token('VIP_INIT', line):
                self.properties['VIP_INIT'] = ' '.join(line.split('!')[0].split()[1:])
            # Find standalone equilibration keywords
            if [i for i in line.split() if i in EQUIL_SINGLE_KEYWORDS]:
                for word in EQUIL_SINGLE_KEYWORDS:
                    if nfo.check_token(word, line):
                        self.properties[word] = ''
            # Handle integrated saturation initialization options
            if [i for i in line.split() if i in EQUIL_INTSAT_KEYWORDS]:
                for key in EQUIL_INTSAT_KEYWORDS:
                    if nfo.check_token(key, line):
                        if nfo.get_token_value(key, line, file_as_list) == 'MOBILE':
                            self.properties[key] = 'MOBILE'
                        else:
                            self.properties[key] = ''

            # Find starting index of an equil-related table. There is usually only one per equil file
            if [i for i in line.split() if i in EQUIL_TABLE_KEYWORDS]:
                for table_keyword in EQUIL_TABLE_KEYWORDS:
                    if nfo.check_token(table_keyword, line):
                        equil_table_indices[table_keyword] = [line_indx + 1, len(file_as_list)]
                        table_being_read = table_keyword
                        start_reading_table = True
                        if table_keyword == 'COMPOSITION' and [i for i in line.split() if i in
                                                               EQUIL_COMPOSITION_OPTIONS]:
                            for comp_key in EQUIL_COMPOSITION_OPTIONS:
                                if nfo.check_token(comp_key, line):
                                    self.properties[comp_key] = float(nfo.get_expected_token_value(comp_key,
                                                                                                   line, file_as_list))
                line_indx += 1
                continue
            # Find ending index of an equil-related table. There is usually only one per equil file
            if start_reading_table and 'DEPTH' not in line.split():
                for potential_endkeyword in EQUIL_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        equil_table_indices[table_being_read][1] = line_indx
                        start_reading_table = False
                        break

            line_indx += 1

        # Read in table if there is one
        for key in equil_table_indices.keys():
            self.properties[key] = nfo.read_table_to_df(file_as_list[
                equil_table_indices[key][0]:equil_table_indices[key][1]])



"""Class to hold Nexus Water properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.Utils.factory_methods import get_empty_dict_union, get_empty_list_nexus_water_params
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Utils.invert_nexus_map import invert_nexus_map
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import WaterUnits


@dataclass  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusWaterParams:
    """Class to represent a single set of water property parameters, i.e., density, compressibility, etc.

    Attributes:
        temperature (Optional[float]): Temperature at which the rest of the water property parameters apply
        salinity (Optional[float]): Salinity at which the rest of the water property parameters apply
        density (Optional[float]): Water density at standard conditions
        compressibility (float): Water compressibility
        formation_volume_factor (float): Water formation volume factor at the given temperature and reference pressure
        viscosity (float): Water viscosity at the given temperature and reference pressure
        viscosity_compressibility (Optional[float]): Water viscosity compressibility.
    """

    # General parameters
    water_compressibility: Optional[float] = None
    water_formation_volume_factor: Optional[float] = None
    water_viscosity: Optional[float] = None
    water_viscosity_compressibility: Optional[float] = None
    temperature: Optional[float] = None
    salinity: Optional[float] = None
    water_density: Optional[float] = None


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusWaterMethod(DynamicProperty):
    """Class to hold Nexus Water properties.

    Attributes:
        file (NexusFile): Nexus water file object
        input_number (int): Water method number in Nexus fcs file
        reference_pressure (float): Reference pressure for BW and, if CVW is present, for VISW
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding properties for generic dynamic method. Defaults to empty dictionary.
        parameters (list[NexusWaterParams]): list of water parameters, such as density, viscosity, etc.
    """

    file: NexusFile
    reference_pressure: Optional[float] = None
    properties: dict[str, Union[str, int, float, Enum, list[str],
                     pd.DataFrame, dict[str, Union[float, pd.DataFrame]]]] \
        = field(default_factory=get_empty_dict_union)
    parameters: list[NexusWaterParams] = field(default_factory=get_empty_list_nexus_water_params)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 reference_pressure: Optional[float] = None,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None,
                 parameters: Optional[list[NexusWaterParams]] = None) -> None:
        self.reference_pressure = reference_pressure
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        if parameters is not None:
            self.parameters = parameters
        else:
            self.parameters = []
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'TEMP': ('temperature', float),
            'PREF': ('reference_pressure', float),
            'DENW': ('water_density', float),
            'CW': ('water_compressibility', float),
            'BW': ('water_formation_volume_factor', float),
            'VISW': ('water_viscosity', float),
            'CVW': ('water_viscosity_compressibility', float)
            }
        return keywords

    @property
    def units(self) -> WaterUnits:
        """Returns the attribute to unit map for the Water method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return WaterUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with water data, in Nexus file format."""
        param_to_nexus_keyword_map = invert_nexus_map(NexusWaterMethod.get_keyword_mapping())

        printable_str = ''
        water_dict = self.properties
        # Print description if present
        if 'DESC' in water_dict.keys() and isinstance(water_dict['DESC'], list):
            for desc_line in water_dict['DESC']:
                printable_str += 'DESC ' + desc_line + '\n'
        if self.reference_pressure is not None:
            printable_str += f'PREF {self.reference_pressure}\n'
        for key, value in water_dict.items():
            if isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key not in ['DESC']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        water_params = self.parameters
        temp_val = None
        sal_val = None
        for i in range(len(water_params)):
            water_param_dict = water_params[i].__dict__
            space_prefix = ''
            if water_param_dict['temperature']:
                if water_param_dict['temperature'] != temp_val:
                    printable_str += f"TEMP {water_param_dict['temperature']}\n"
                    temp_val = water_param_dict['temperature']
                space_prefix += '    '
            if water_param_dict['salinity']:
                if water_param_dict['salinity'] != sal_val:
                    printable_str += f"{space_prefix}SALINITY {water_param_dict['salinity']}\n"
                    sal_val = water_param_dict['salinity']
                space_prefix += '    '
            for key in param_to_nexus_keyword_map.keys():
                if key in water_param_dict and water_param_dict[key] is not None and key != 'temperature':
                    printable_str += f'{space_prefix}{param_to_nexus_keyword_map[key]} {water_param_dict[key]}\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus Water file contents and populate NexusWaterMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize properties
        water_props_dict_none: dict[str, Optional[float]] = {'DENW': None, 'CW': None, 'BW': None,
                                                             'VISW': None, 'CVW': None}
        water_props_dict: dict[str, Optional[float]] = water_props_dict_none.copy()

        line_indx = 0
        reading_props = False  # Flag indicating if actively reading a property section
        found_params = False  # Flag inidicating if identified any WATER keywords
        temp: Optional[float] = None  # Variable to hold temperature for property section
        sal: Optional[float] = None  # Variable to hold salinity for property section
        for line in file_as_list:

            # Read in reference pressure
            if nfo.check_token('PREF', line):  # Reference pressure
                self.reference_pressure = float(nfo.get_expected_token_value('PREF', line, file_as_list))

            # Check if reading properties
            if nfo.check_token('CW', line):
                reading_props = True
                found_params = True

            # If found a new property section based on TEMP or SALINITY repeat
            if reading_props and (nfo.check_token('TEMP', line) or nfo.check_token('SALINITY', line)):
                reading_props = False
                # Append new parameters to list of water parameters
                params = NexusWaterParams(temperature=temp,
                                          salinity=sal,
                                          water_density=water_props_dict['DENW'],
                                          water_compressibility=water_props_dict['CW'],
                                          water_formation_volume_factor=water_props_dict['BW'],
                                          water_viscosity=water_props_dict['VISW'],
                                          water_viscosity_compressibility=water_props_dict['CVW']
                                          )
                self.parameters.append(params)
                # Reset property dictionary
                water_props_dict = water_props_dict_none.copy()  # Reset

            # Read in relevant water properties in line
            if nfo.check_token('TEMP', line):
                temp = float(nfo.get_expected_token_value('TEMP', line, file_as_list))
            if nfo.check_token('SALINITY', line):
                sal = float(nfo.get_expected_token_value('SALINITY', line, file_as_list))
            for key in water_props_dict.keys():
                if nfo.check_token(key, line):
                    found_params = True
                    water_props_dict[key] = float(nfo.get_expected_token_value(key, line, file_as_list))

            line_indx += 1

        if found_params:
            # Append new parameters to list of water parameters
            params = NexusWaterParams(temperature=temp,
                                      salinity=sal,
                                      water_density=water_props_dict['DENW'],
                                      water_compressibility=water_props_dict['CW'],
                                      water_formation_volume_factor=water_props_dict['BW'],
                                      water_viscosity=water_props_dict['VISW'],
                                      water_viscosity_compressibility=water_props_dict['CVW']
                                      )
            self.parameters.append(params)



"""Class that is used to represent a nexus node in the network."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Node import Node


@dataclass(kw_only=True, repr=False)
class NexusNode(Node):
    """Class that is used to represent a nexus node in the network."""
    x_pos: Optional[float] = None
    y_pos: Optional[float] = None
    number: Optional[int] = None
    temp: Optional[float] = None
    station: Optional[str] = None

    def __init__(self, properties_dict: dict[str, None | int | str | float]) -> None:
        # call the init of the DataObjectMixin
        super(Node, self).__init__({})
        for key, prop in properties_dict.items():
            self.__setattr__(key, prop)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords = {
            'NAME': ('name', str),
            'TYPE': ('type', str),
            'DEPTH': ('depth', float),
            'TEMP': ('temp', float),
            'X': ('x_pos', float),
            'Y': ('y_pos', float),
            'NUMBER': ('number', int),
            'STATION': ('station', str),
        }
        return keywords

    def update(self, input_dictionary:  dict[str, None | float | int | str]) -> None:
        """Updates a node based on a dictionary of attributes."""
        for k, v in input_dictionary.items():
            if v is None:
                continue
            if hasattr(self, '_NexusNode__' + k):
                setattr(self, '_NexusNode__' + k, v)
            elif hasattr(super(), '_Node__' + k):
                setattr(self, '_Node__' + k, v)



from __future__ import annotations
from enum import Enum
from typing import TYPE_CHECKING, Union
from uuid import UUID

import pandas as pd

if TYPE_CHECKING:
    from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
    from ResSimpy.Nexus.DataModels.NexusWaterMethod import NexusWaterParams


# Factory methods for generating empty lists with typing
def get_empty_list_str() -> list[str]:
    value: list[str] = []
    return value


# Factory method for generating empty dictionary with typing
def get_empty_dict_union() -> dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                   dict[str, Union[float, pd.DataFrame]]]]:
    value: dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame, dict[str, Union[float, pd.DataFrame]]]] = {}
    return value


# Factory method for generating empty dictionary for eos options
def get_empty_eosopt_dict_union() -> \
        dict[str, Union[str, int, float, pd.DataFrame, list[str], dict[str, float], tuple[str, dict[str, float]], dict[
            str, pd.DataFrame]]]:
    value: dict[str, Union[
        str, int, float, pd.DataFrame, list[str], dict[str, float], tuple[str, dict[str, float]], dict[
            str, pd.DataFrame]]] = {}
    return value


# Factory method for generating empty dictionary for hysteresis parameters
def get_empty_hysteresis_dict() -> dict[str, Union[str, float, dict[str,
                                        Union[str, float, dict[str, Union[str, float]]]]]]:
    value: dict[str, Union[str, float, dict[str, Union[str, float, dict[str, Union[str, float]]]]]] = {}
    return value


def get_empty_list_str_nexus_file() -> list[Union[str, NexusFile]]:
    value: list[Union[str, NexusFile]] = []
    return value


def get_empty_list_nexus_file() -> list[NexusFile]:
    value: list[NexusFile] = []
    return value


def get_empty_dict_int_nexus_file() -> dict[int, NexusFile]:
    value: dict[int, NexusFile] = {}
    return value


def get_empty_dict_uuid_list_int() -> dict[UUID, list[int]]:
    value: dict[UUID, list[int]] = {}
    return value


def get_empty_list_nexus_water_params() -> list[NexusWaterParams]:
    value: list[NexusWaterParams] = []
    return value



"""Class representing a single Wellhead in a Nexus Network."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Wellhead import Wellhead


@dataclass(kw_only=True)
class NexusWellhead(Wellhead):
    """Class representing a single Wellhead in a Nexus Network.

    Attributes:
        well (str):  Associates the wellhead to the well. (WELL)
        name (str):  The name of the wellhead. (NAME)
        type (str):  The type of well. (TYPE)
        depth (float):  The depth of the wellhead  (DEPTH)
        x_pos (float):  The x-coordinate location of the wellhead  (X)
        y_pos (float):  The y-coordinate location of the wellhead  (Y)
        pvt_method (int):  The PVT table numbers to be used for the well and wellhead connections. (IPVT)
        water_method (int):  The Water PVT table numbers to be used for the well and wellhead connections. (IWAT)
        bat_method (int):  The separator battery numbers associated with the well and wellhead connections. (IBAT)
        measured_depth_in (float):  The measured depth at the start of the well interval  (MDIN)
        measured_depth_out (float):  The measured depth at the end of the well interval  (MDOUT)
        hyd_method (int):  The hydraulic method used. (METHOD)
        number (int):  The well number. (NUMBER)
        diameter (float):  The wellbore diameter  (DIAM)
        inner_diameter (float):  The well inner diameter  (INNERDIAM)
        roughness (float):  The well roughness. (ROUGHNESS)
        length (float):  The length of the well  (LENGTH)
        temperature (float):  The temperature of the fluid in the well  (TEMP)
        elevation_profile (str):  The well elevation profile. (ELEVPR)
        temperature_profile (str):  The well temperature profile. (TEMPPR)
        dp_add (float):  The additional pressure drop across the well  (DPADD)
        rate_mult (float):  The rate multiplier for the well. (RATEMULT)
        delta_depth (float):  The depth difference between the two points in the connection  (DDEPTH)
        heat_transfer_coeff (float):  The heat transfer coefficient for the well  (HTC)
        dt_add (float):  The additional temperature difference across the well  (DTADD).
    """

    pvt_method: Optional[int] = None
    water_method: Optional[int] = None
    bat_method: Optional[int] = None
    measured_depth_in: Optional[float] = None
    measured_depth_out: Optional[float] = None
    hyd_method: Optional[int] = None
    number: Optional[int] = None
    diameter: Optional[float] = None
    inner_diameter: Optional[float] = None
    roughness: Optional[float] = None
    length: Optional[float] = None
    temperature: Optional[float] = None
    elevation_profile: Optional[str] = None
    temperature_profile: Optional[str] = None
    dp_add: Optional[float] = None
    rate_mult: Optional[float] = None
    delta_depth: Optional[float] = None
    heat_transfer_coeff: Optional[float] = None
    dt_add: Optional[float] = None

    def __init__(self, properties_dict: dict[str, None | int | str | float]) -> None:
        super(Wellhead, self).__init__({})
        for key, prop in properties_dict.items():
            self.__setattr__(key, prop)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'WELL': ('well', str),
            'NAME': ('name', str),
            'TYPE': ('wellhead_type', str),
            'DEPTH': ('depth', float),
            'X': ('x_pos', float),
            'Y': ('y_pos', float),
            'IPVT': ('pvt_method', int),
            'IWAT': ('water_method', int),
            'IBAT': ('bat_method', int),
            'MDIN': ('measured_depth_in', float),
            'MDOUT': ('measured_depth_out', float),
            'METHOD': ('hyd_method', int),
            'NUMBER': ('number', int),
            'DIAM': ('diameter', float),
            'INNERDIAM': ('inner_diameter', float),
            'ROUGHNESS': ('roughness', float),
            'LENGTH': ('length', float),
            'TEMP': ('temperature', float),
            'ELEVPR': ('elevation_profile', str),
            'TEMPPR': ('temperature_profile', str),
            'DPADD': ('dp_add', float),
            'RATEMULT': ('rate_mult', float),
            'DDEPTH': ('delta_depth', float),
            'HTC': ('heat_transfer_coeff', float),
            'DTADD': ('dt_add', float),
        }

        return nexus_mapping



from dataclasses import dataclass
from abc import ABC


@dataclass(kw_only=True)
class Rock(ABC):
    """The abstract base class for a collection of rock property inputs.

    Attributes:
        inputs (dict[int, DynamicProperty]): Collection of rock property inputs, as a dictionary.
    """

    @property
    def inputs(self):
        raise NotImplementedError("Implement this in the derived class")



# A list of valid Nexus Tokens. Please add to this as you find / use them.
from ResSimpy.Nexus.NexusKeywords.aquifer_keywords import AQUIFER_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.equil_keywords import EQUIL_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.fcs_keywords import FCS_ALL_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.gaslift_keywords import GASLIFT_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.options_keywords import OPTIONS_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.relpm_keywords import RELPM_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.rock_keywords import ROCK_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.runcontrol_keywords import RUNCONTROL_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.separator_keywords import SEPARATOR_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_KEYWORDS, STRUCTURED_GRID_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.surface_keywords import SURFACE_KEYWORDS, CONNECTION_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.valve_keywords import VALVE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.water_keywords import WATER_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.wells_keywords import WELLS_KEYWORDS

GENERAL_KEYWORDS = ['DESC', 'LABEL']

RELPERM_ENDPT_KEYWORDS = ['SWL', 'SWR', 'SWU', 'SGL', 'SGR', 'SGU', 'SWRO', 'SGRO', 'SGRW', 'KRW_SWRO', 'KRW_SWU',
                          'KRG_SGRO', 'KRG_SGU', 'KRO_SWL', 'KRO_SWR', 'KRO_SGL', 'KRO_SGR', 'KRW_SGL', 'KRW_SGR',
                          'KRG_SGRW', 'SGTR', 'SOTR']

UNIT_KEYWORDS = ['ENGLISH', 'METRIC', 'METKG/CM2', 'METBAR', 'LAB', 'FAHR', 'CELSIUS', 'KELVIN', 'RANKINE']

# Keywords that sometimes follow other keywords before the value
INTERMEDIATE_KEYWORDS = ['SET', 'METHOD', 'NETWORK']  # always followed by an int, then a file path

# Keywords that require another keyword after their declaration
STARTING_KEYWORDS = ['RESTART']

# Combine all lists into the complete list
VALID_NEXUS_KEYWORDS = GENERAL_KEYWORDS + RUNCONTROL_KEYWORDS + UNIT_KEYWORDS \
                       + INTERMEDIATE_KEYWORDS + STARTING_KEYWORDS + RELPERM_ENDPT_KEYWORDS + PVT_KEYWORDS \
                       + FCS_ALL_KEYWORDS + WELLS_KEYWORDS + WATER_KEYWORDS + VALVE_KEYWORDS + SURFACE_KEYWORDS + \
                       CONNECTION_KEYWORDS + GRID_ARRAY_KEYWORDS + STRUCTURED_GRID_KEYWORDS + SEPARATOR_KEYWORDS + \
                       ROCK_KEYWORDS + AQUIFER_KEYWORDS + EQUIL_KEYWORDS + GASLIFT_KEYWORDS + HYD_KEYWORDS + \
                       OPTIONS_KEYWORDS + PVT_KEYWORDS + RELPM_KEYWORDS + WELLS_KEYWORDS



# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

OPTIONS_KEYWORDS = ['ALL', 'AQUIFER', 'AUTO', 'CENTER', 'DATUM', 'DEF_DATUM', 'DESC', 'DP',
                    'ENDGLOBAL_METHOD_OVERRIDES', 'ENDGRIDTOPROC', 'ENDREGBLK', 'ENDREGDATA', 'ENDSUBREGIONS',
                    'EQ_LIST', 'FLEXDP', 'GLOBAL_METHOD_OVERRIDES', 'GRID', 'GRIDBLOCKS', 'GRIDTOPROC', 'INTSAT',
                    'NAME', 'NOCHK_SAL_TEMP', 'NONEQ', 'NUMBER', 'OFF', 'OVERREAD', 'PATTERN', 'PC', 'PD',
                    'POROSITY_INDEPENDENCE', 'PROCESS', 'PSTD', 'REGBLK', 'REGDATA', 'RESERVOIR', 'RES_TEMP', 'SOMOPT1',
                    'STONE1', 'SUBREGION', 'SUBREGIONS', 'SW', 'TSTD', 'VIP_INJ_PERF_KR_SCALING']



"""Class for dealing with logfiles produced from the nexus simulation engine."""
from __future__ import annotations

import os
from datetime import datetime
from typing import Optional, TYPE_CHECKING
import ResSimpy.Nexus.nexus_file_operations as nfo

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


class Logging:
    def __init__(self, model: NexusSimulator) -> None:
        """Class for controlling all logging and logfile (*.log) related functionality.

        Args:
            model: NexusSimulator instance
            __job_id (int): Run job ID for executed runs
            __simulation_start_time (Optional[str]): Execution start time of the simulation when submitted \
                to calculation engine
            __simulation_end_time (Optional[str]): Execution end time of the last time the simulation was run
            __previous_run_time (Optional[str]): Difference between simulation execution start time and end time.
        """
        self.__model = model
        self.__job_id: int = -1
        self.__simulation_start_time: Optional[str] = None
        self.__simulation_end_time: Optional[str] = None
        self.__previous_run_time: Optional[str] = None

    @staticmethod
    def get_simulation_time(line: str) -> str:
        """Searches for the simulation time in a line.

        Args:
            line (str): line to search for the simulation time

        Raises:
            ValueError: Throws error if get_next_value doesn't find any subsequent value in the line

        Returns:
            str: value found after TIME card in a line
        """
        value_found = False
        value = ''
        line_string = line
        while value_found is False:
            next_value = nfo.get_next_value(0, [line_string], line_string)
            if next_value is None:
                raise ValueError(
                    f'No next value found in the line supplied, line: {line_string}')
            if next_value == 'on':
                line_string = line_string.replace(next_value, '', 1)
                next_value = nfo.get_next_value(0, [line_string], line_string)
                if next_value is None:
                    raise ValueError(
                        f'No next value found in the line supplied, line: {line_string}')
                for c in range(6):
                    line_string = line_string.replace(next_value, '', 1)
                    next_value = nfo.get_next_value(0, [line_string], line_string)
                    if next_value is None:
                        raise ValueError(
                            f'No next value found in the line supplied, line: {line_string}')
                    value += next_value + (' ' if c < 5 else '')
                value_found = True
            line_string = line_string.replace(next_value, '', 1)
        return value

    @staticmethod
    def convert_server_date(original_date: str) -> datetime:
        """Convert a datetime string from the server for when the simulation was started to a strptime object.

        Args:
            original_date (str): string of a date with format: "Mon Jan 01 00:00:00 CST 2020"

        Returns:
            datetime: datetime object derived from the input string
        """

        date_format = '%a %b %d %X %Z %Y'
        converted_date = original_date

        # Convert CDT and CST timezones as Python doesn't work with CDT for some reason
        if 'CDT' in original_date:
            converted_date = converted_date.replace('CDT', '-0500', 1)
            date_format = '%a %b %d %X %z %Y'
        elif 'CST' in original_date:
            converted_date = converted_date.replace('CST', '-0600', 1)
            date_format = '%a %b %d %X %z %Y'

        return datetime.strptime(converted_date, date_format)

    @staticmethod
    def get_errors_warnings_string(log_file_line_list: list[str]) -> Optional[str]:
        """Retrieves the number of warnings and errors from the simulation log output,
        and formats them as a string.

        Args:
            log_file_line_list (list[str]): log file formatted as a list of strings with \
                a new list entry per line

        Returns:
            Optional[str]: string containing the errors and warnings from the simulation log. \
                None if error/warning set is too short
        """
        error_warning = ""
        for line in log_file_line_list:
            modified_line = line.lower()
            if "errors" in modified_line and "warnings" in modified_line:
                error_warning = modified_line

        error_warning_list = [x for x in error_warning.split(" ") if x != ""]

        error_warning_list = [nfo.clean_up_string(x) for x in error_warning_list]

        if len(error_warning_list) < 4:
            return None

        errors = error_warning_list[1]
        warnings = error_warning_list[3]

        error_warning_str = f"Simulation complete - Errors: {errors} and Warnings: {warnings}"
        return error_warning_str

    def get_simulation_start_time(self) -> str:
        """Get the start time of an executed simulation run, if no simulation start time returns '-'."""
        self.get_simulation_status()
        if self.__simulation_start_time is not None:
            return self.__simulation_start_time
        else:
            return '-'

    def get_simulation_end_time(self) -> str:
        """Get the end time of an executed simulation run if it has completed, if no simulation end time returns '-'."""
        self.get_simulation_status()
        if self.__simulation_end_time is not None:
            return self.__simulation_end_time
        else:
            return '-'

    def get_job_id(self) -> int:
        """Get the job Id of a simulation run."""
        return self.__job_id

    def __get_log_path(self, from_startup: bool = False) -> Optional[str]:
        """Returns the path of the log file for the simulation.

        Args:
            from_startup (bool, optional): Searches the same directory as the original_fcs_file_path if True. \
            Otherwise searches the destination folder path, failing this then searches the \
            original_fcs_file_path if False. Defaults to False.

        Returns:
            Optional[str]: The path of the .log file from the simulation if found. If not found returns None.
        """
        folder_path = os.path.dirname(
            self.__model.original_fcs_file_path) if from_startup else os.path.dirname(self.__model.origin)
        files = os.listdir(folder_path)
        original_fcs_file_location = os.path.basename(self.__model.original_fcs_file_path)
        log_file_name = os.path.splitext(original_fcs_file_location)[
                            0] + ".log" if from_startup else self.__model.root_name + ".log"

        if log_file_name in files:
            if from_startup:
                file_location = folder_path
            else:
                file_location = self.__model.destination if self.__model.destination is not None else folder_path

            log_file_path = file_location + "/" + log_file_name
            return log_file_path
        else:
            return None

    def __update_simulation_start_and_end_times(self, log_file_line_list: list[str]) -> None:
        """Updates the stored simulation execution start and end times from the log files.

        Args:
            log_file_line_list (list[str]): log file information represented with a new entry per line of the file.
        """
        for line in log_file_line_list:
            if nfo.check_token('start generic pdsh   prolog', line):
                value = self.get_simulation_time(line)
                self.__simulation_start_time = value

            if nfo.check_token('end generic pdsh   epilog', line):
                value = self.get_simulation_time(line)
                self.__simulation_end_time = value

    def get_simulation_status(self, from_startup: bool = False) -> Optional[str]:
        """Gets the run status of the latest simulation run.

        Args:
            from_startup (bool, optional): Searches the same directory as the original_fcs_file_path if True. \
            Otherwise searches the destination folder path, failing this then searches the \
            original_fcs_file_path if False. Defaults to False.

        Raises:
            NotImplementedError: If log file is not found - only supporting simulation status from log files

        Returns:
            Optional[str]: the error/warning string if the simulation has finished, otherwise \
                returns the running job ID. Empty string if a logfile is not found and from_start up is True
        """
        log_file = self.__get_log_path(from_startup)
        if log_file is None:
            if from_startup:
                return ''
            raise NotImplementedError(
                "Only retrieving status from a log file is supported at the moment")
        else:
            log_file_line_list = nfo.load_file_as_list(log_file)
            self.__update_simulation_start_and_end_times(log_file_line_list)
            job_finished = 'Nexus finished\n' in log_file_line_list
            if job_finished:
                self.__previous_run_time = self.__get_start_end_difference() if from_startup \
                    else self.__previous_run_time
                return self.get_errors_warnings_string(log_file_line_list=log_file_line_list)
            else:
                job_number_line = [
                    x for x in log_file_line_list if 'Job number:' in x]
                if len(job_number_line) > 0:
                    self.__job_id = int(job_number_line[0].split(":")[1])
                    return f"Job Running, ID: {self.__job_id}"
        return None

    def __get_start_end_difference(self) -> Optional[str]:
        """Returns a string with the previous time taken when the base case was run.

        Returns:
            Optional[str]: returns a human readable string of how long the simulation took to run
        """
        if self.__simulation_start_time is None or self.__simulation_end_time is None:
            return None

        start_date = self.convert_server_date(self.__simulation_start_time)
        end_date = self.convert_server_date(self.__simulation_end_time)

        total_difference = (end_date - start_date)
        days = int(total_difference.days)
        hours = int(total_difference.seconds / (60 * 60))
        minutes = int((total_difference.seconds / 60) - (hours * 60))
        seconds = int(total_difference.seconds -
                      (hours * 60 * 60) - (minutes * 60))

        return f"{days} Days, {hours} Hours, {minutes} Minutes {seconds} Seconds"

    def get_base_case_run_time(self) -> str:
        """Get the time taken for the base case to run. Returns '-' if no run time found."""
        if self.__previous_run_time is not None:
            return self.__previous_run_time
        else:
            return '-'

    def get_simulation_progress(self) -> float:
        """Returns the simulation progress from log files.

        Raises:
            NotImplementedError: Only retrieving status from a log file is supported at the moment
            ValueError: if no times from the runcontrol file are read in

        Returns:
            Optional[float]: how long through a simulation run as a proportion of the number of days \
                simulated as stated in the runcontrol
        """
        log_file_path = self.__get_log_path()
        if log_file_path is None:
            raise NotImplementedError("Only retrieving status from a log file is supported at the moment")
        log_file = nfo.load_file_as_list(log_file_path)

        read_in_times = False
        time_heading_location = None
        last_time = None
        for line in log_file:
            case_name_string = f"Case Name = {self.__model.root_name}"
            if case_name_string in line:
                read_in_times = True
                continue
            if read_in_times and nfo.check_token('TIME', line):
                heading_location = 0
                line_string = line
                while len(line_string) > 0:
                    next_value = nfo.get_next_value(0, [line_string], line_string)
                    if next_value is None:
                        break

                    line_string = line_string.replace(next_value, '', 1)
                    if next_value == 'TIME':
                        time_heading_location = heading_location
                    heading_location += 1

            if read_in_times and time_heading_location is not None:
                line_string = line
                next_value = nfo.get_next_value(0, [line_string], line_string)
                if next_value is not None and next_value.replace('.', '', 1).isdigit():
                    if time_heading_location == 0 and (last_time is None or float(next_value) > float(last_time)):
                        last_time = next_value
                    for x in range(time_heading_location):
                        line_string = line_string.replace(next_value, '', 1)
                        next_value = nfo.get_next_value(0, [line_string], line_string)
                        if next_value is None:
                            break
                        # When we reach the time column, read in the time value.
                        if x == (time_heading_location - 1) and \
                                (last_time is None or float(next_value) > float(last_time)):
                            last_time = next_value

        if last_time is not None:
            days_completed = self.__model._sim_controls.convert_date_to_number(last_time)
            if self.__model._sim_controls.times is None:
                raise ValueError("No times provided in the instance - please read them in from runcontrol file")
            total_days = self.__model._sim_controls.convert_date_to_number(self.__model._sim_controls.times[-1])
            return round((days_completed / total_days) * 100, 1)

        return 0



# Currently ignoring functional form rel perm inputs, only tabular keywords are handled.
# Also currently ignoring spline options

RELPM_TABLE_KEYWORDS = ['WOTABLE', 'GOTABLE', 'GWTABLE', 'WOTABLE_IMB', 'GOTABLE_IMB', 'GWTABLE_IMB',
                        'WAGTABLE_WIMB', 'GRMTABLE_SGTR', 'GRMTABLE', 'GDWTABLE', 'LOW_SALTAB',
                        'PRSTAB', 'WAGTEST']
RELPM_SINGLE_KEYWORDS = ['KROINT', 'STONE2', 'STONE1', 'SOMOPT1', 'SOMOPT3', 'LOW_SAL', 'GW3PHASE',
                         'NOCHK', 'VIP_INJ_PERF_KR_SCALING', 'VEGO', 'VEGO_PC', 'VEWO', 'VEWO_PC',
                         'IRELPM_LOWIFT', 'KRWINT', 'STONE2_WAT', 'STONE1_WAT', 'SWMOPT1',
                         'FREEZE_PCGO', 'FREEZE_PCWO', 'FREEZE_PCGW', 'LOW_SAL_IRR', 'VIP_RELPM']
RELPM_HYSTERESIS_PRIMARY_KEYWORDS = ['NONE', 'KRW', 'KRG', 'KROW', 'PCWO', 'PCGO', 'WAG',
                                     'TOLREV', 'TOLHYS', 'NOCHK_HYS']
RELPM_KEYWORDS_VALUE_FLOAT = ['SOMOPT2', 'ST1EXP', 'ST1EPS', 'ST1NU', 'GRMDSW', 'PERMBASIS', 'PORBASIS', 'DRELPM',
                              'TENTHR', 'XEX', 'TENI', 'SWMOPT2']
RELPM_KEYWORDS_VALUE_STR = ['HYSTERESIS', 'JFUNC', 'SCALING', 'SCALING_PC', 'IFT', 'NEARCRIT', 'DERIVATIVES',
                            'RECONSTRUCT']
RELPM_NONDARCY_KEYWORDS = ['NONDARCY_GAS', 'NONDARCY_OIL']
RELPM_NONDARCY_PARAMS = ['BETA', 'BETA0', 'BETA1', 'BETA2', 'BETA3', 'BETA4', 'BETA5', 'IFT_THRES']

RELPM_KEYWORDS = RELPM_TABLE_KEYWORDS + RELPM_SINGLE_KEYWORDS + RELPM_KEYWORDS_VALUE_FLOAT + \
                 RELPM_KEYWORDS_VALUE_STR + RELPM_NONDARCY_KEYWORDS



from __future__ import annotations

import warnings
from datetime import datetime, timedelta
from functools import cmp_to_key
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Nexus.constants import DATE_WITH_TIME_LENGTH


class SimControls:
    def __init__(self, model) -> None:
        """Class for controlling all runcontrol and time related functionality
        Args:
            model: NexusSimulator instance
            __times (None | list[str]): list of times to be included in the runcontrol file
            __date_format_string (str): How the dates should formatted based on date_format.

        """
        self.__model = model
        self.__times: None | list[str] = None
        self.__date_format_string: str = ''

    @property
    def date_format_string(self):
        return self.__date_format_string

    @date_format_string.setter
    def date_format_string(self, value):
        self.__date_format_string = value

    @property
    def times(self):
        return self.__times if self.__times is not None else []

    @staticmethod
    def get_times(times_file: list[str]) -> list[str]:
        """Retrieves a list of TIMES from the supplied Runcontrol / Include file.

        Args:
        ----
            times_file (list[str]): list of strings with each line from the file a new entry in the list
        Returns:
            list[str]: list of all the values following the TIME keyword in supplied file, \
                empty list if no values found
        """
        times = []
        for line in times_file:
            if nfo.check_token('TIME', line):
                value = nfo.get_token_value('TIME', line, times_file)
                if value is not None:
                    times.append(value)

        return times

    @staticmethod
    def delete_times(file_content: list[str]) -> list[str]:
        """Deletes times from file contents.

        Args:
            file_content (list[str]):  list of strings with each line from the file a new entry in the list.

        Returns:
            list[str]: the modified file without any TIME cards in
        """
        new_file: list[str] = []
        previous_line_is_time = False
        for line in file_content:
            if "TIME " not in line and (previous_line_is_time is False or line != '\n'):
                new_file.append(line)
                previous_line_is_time = False
            elif "TIME " in line:
                previous_line_is_time = True
            else:
                previous_line_is_time = False
        return new_file

    @staticmethod
    def remove_times_from_file(file_content: list[str], output_file_path: str) -> None:
        """Removes the times from a file - used for replacing with new times
        Args:
            file_content (list[str]): a list of strings containing each line of the file as a new entry
            output_file_path (str): path to the file to output to.
        """
        new_file_content = SimControls.delete_times(file_content)

        new_file_str = "".join(new_file_content)

        with open(output_file_path, "w") as text_file:
            text_file.write(new_file_str)

    def convert_date_to_number(self, date: str | float) -> float:
        """Converts a date to a number designating number of days from the start date.

        Args:
        ----
            date (str | float): a date or time stamp from a Nexus simulation
        Raises:
            ValueError: if supplied incorrect type for 'date' parameter

        Returns:
        -------
            float: the difference between the supplied date and the start date of the simulator
        """
        # If we can retrieve a number of days from date, use that, otherwise convert the string date to a number of days
        try:
            converted_date: str | float = float(date)
        except ValueError:
            if not isinstance(date, str):
                raise ValueError("convert_date_to_number: Incorrect type for 'date' parameter")
            converted_date = date

        if isinstance(converted_date, float):
            date_format = self.date_format_string
            if len(self.__model.start_date) == DATE_WITH_TIME_LENGTH:
                date_format += "(%H:%M:%S)"
            start_date_as_datetime = datetime.strptime(self.__model.start_date, date_format)
            date_as_datetime = start_date_as_datetime + timedelta(days=converted_date)
        else:
            start_date_format = self.date_format_string
            if len(self.__model.start_date) == DATE_WITH_TIME_LENGTH:
                start_date_format += "(%H:%M:%S)"
            end_date_format = self.date_format_string
            if len(converted_date) == DATE_WITH_TIME_LENGTH:
                end_date_format += "(%H:%M:%S)"
            date_as_datetime = datetime.strptime(converted_date, end_date_format)
            start_date_as_datetime = datetime.strptime(self.__model.start_date, start_date_format)

        difference = date_as_datetime - start_date_as_datetime
        return difference.total_seconds() / timedelta(days=1).total_seconds()

    def compare_dates(self, x: str | float, y: str | float) -> int:
        """Comparator for two supplied dates or numbers.

        Args:
        ----
            x (str | float): first date to compare
            y (str | float): second date to compare

        Returns:
        -------
            int: -1 if y > x, 0 if y == x, 1 if y < x
        """
        date_comp = self.convert_date_to_number(x) - self.convert_date_to_number(y)
        if date_comp < 0:
            date_comp_int = -1
        elif date_comp == 0:
            date_comp_int = 0
        else:
            date_comp_int = 1

        return date_comp_int

    def sort_remove_duplicate_times(self, times: list[str]) -> list[str]:
        """Removes duplicates and nans from the times list, then sorts them.

        Args:
        ----
            times (list[str]): list of times to remove duplicates from

        Returns:
        -------
            list[str]: list of times without duplicates
        """
        new_times = []
        for i in times:
            i_value = i.strip()
            if i != i or i_value in new_times:
                continue
            new_times.append(i_value)
        new_times = sorted(new_times, key=cmp_to_key(self.compare_dates))
        return new_times

    def check_date_format(self, date: str | float) -> None:
        """Checks that a supplied date is in the correct format.

        Args:
        ----
            date (str | float): date to check the format of
        Raises:
            ValueError: If a date provided isn't in a date format that the model expects
        """
        try:
            float(date)
        except ValueError:
            # Value isn't a float - attempt to extract date from value
            try:
                date_format = self.date_format_string
                if len(str(date)) == DATE_WITH_TIME_LENGTH:
                    date_format += "(%H:%M:%S)"
                datetime.strptime(str(date), date_format)
            except ValueError:
                current_date_format = self.get_date_format(self.__model.date_format)
                raise ValueError(
                    "Invalid date format " + str(date) + " the model is using " + current_date_format + " date format.")

    @staticmethod
    def get_date_format(date_format: DateFormat) -> str:
        """Returns the date format being used by the model
        formats used: ('MM/DD/YYYY', 'DD/MM/YYYY').
        """

        if date_format is DateFormat.MM_DD_YYYY:
            return 'MM/DD/YYYY'
        else:
            return 'DD/MM/YYYY'

    def __update_times_in_file(self) -> None:
        """Updates the list of times in the Runcontrol file to the current stored values in __times.

        Returns:
            None: writes out a file at the same path as the existing runcontrol file
        """
        self.__model.check_output_path()
        if self.__model.model_files.runcontrol_file is None or \
                self.__model.model_files.runcontrol_file.location is None:
            raise ValueError(f"No file path found for {self.__model.model_files}")
        file_content = self.__model.model_files.runcontrol_file.get_flat_list_str_file
        filename = self.__model.model_files.runcontrol_file.location

        new_file_content = self.__model._sim_controls.delete_times(file_content)

        time_list = self.times
        stop_string = 'STOP\n'
        if stop_string in new_file_content:
            new_file_content.remove(stop_string)

        def prepend_time(s: str) -> str:
            return "TIME " + s

        time_list = [prepend_time(x) for x in time_list]
        num_times = len(time_list)

        new_line_list = ["\n\n"] * num_times
        zipped_list = list(zip(time_list, new_line_list))
        flat_list = [item for sublist in zipped_list for item in sublist]
        flat_list.append(stop_string)
        new_file_content.extend(flat_list)
        new_file_str = "".join(new_file_content)

        with open(filename, "w") as text_file:
            text_file.write(new_file_str)

    def load_run_control_file(self):
        """Loads the run control information into the class instance. \
            If the write_times attribute is True then it expands out any INCLUDE files with the times found within
        Raises:
            ValueError: if the run_control_file attribute is None.
        """
        if self.__model.model_files.runcontrol_file is None:
            warnings.warn(f"Run control file path not found for {self.__model.model_files.location}")
            return
        run_control_file_content = self.__model.model_files.runcontrol_file.get_flat_list_str_file

        if (run_control_file_content is None) or (self.__model.model_files.runcontrol_file.location is None):
            raise ValueError(f"No file path provided for {self.__model.model_files.runcontrol_file.location=}")

        # set the start date
        for line in run_control_file_content:
            if nfo.check_token('START', line):
                value = nfo.get_expected_token_value('START', line, run_control_file_content)
                if value is not None:
                    self.__model.start_date = value

        times = []
        run_control_times = self.get_times(run_control_file_content)
        times.extend(run_control_times)
        if self.__model.start_date is None or self.__model.start_date == '':
            try:
                self.__model.start_date = times[0]
            except IndexError:
                for line in run_control_file_content:
                    if nfo.check_token('TIME', line):
                        value = nfo.get_expected_token_value('TIME', line, run_control_file_content)
                        self.__model.start_date = value
                        warnings.warn(f'Setting start date to first time card found in the runcontrol file as: {value}')
                        break
                warnings.warn('No value found for start date explicitly with START or TIME card')

        self.__times = self.sort_remove_duplicate_times(times)

        # If we don't want to write the times, return here.
        if not self.__model.write_times:
            return
        if self.__model.model_files.runcontrol_file.include_locations is None:
            warnings.warn(f'No includes files found in {self.__model.model_files.runcontrol_file.location}')
            return
        for file in self.__model.model_files.runcontrol_file.include_locations:
            if self.__model.destination is not None:
                self.remove_times_from_file(run_control_file_content, file)

        self.modify_times(content=times, operation='replace')

    def modify_times(self, content: None | list[str] = None, operation: str = 'merge'):
        """Modifies the output times in the simulation.

        Args:
        ----
            content (list[str]], optional): The content to modify using the above operation, \
            represented as a list of strings with a new entry per line of the file. Defaults to None.
            operation (str, optional): operation to perform on the content provided (e.g. 'merge'). Defaults to 'merge'.

        Raises:
        ------
            ValueError: if the supplied dates are before the start date of the simulation
        """
        if content is None:
            content = []
        for time in content:
            self.check_date_format(time)

        new_times = self.sort_remove_duplicate_times(content)
        if len(new_times) > 0 > self.compare_dates(new_times[0], self.__model.start_date):
            raise ValueError(
                f"The supplied date of {new_times[0]} precedes the start date of {self.__model.start_date}")
        operation = operation.lower()
        self.__times = self.__times if self.__times is not None else []

        if operation == 'merge':
            self.__times.extend(content)
        elif operation == 'replace':
            self.__times = content
        elif operation == 'reset':
            self.__times = []
        elif operation == 'remove':
            for time in content:
                if time in self.__times:
                    self.__times.remove(time)

        self.__times = self.sort_remove_duplicate_times(self.__times)

        if self.__model.destination is not None:
            self.__update_times_in_file()



"""Class for representing a well in Nexus. Consists of a list of completions and a well name."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Sequence, Union
from uuid import UUID

from ResSimpy.Nexus.DataModels.NexusCompletion import NexusCompletion
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Utils.generic_repr import generic_repr, generic_str
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Well import Well


@dataclass
class NexusWell(Well):
    __completions: list[NexusCompletion]

    def __init__(self, well_name: str, completions: Sequence[NexusCompletion], unit_system: UnitSystem) -> None:
        if not isinstance(completions, list):
            completions = list(completions)
        self.__completions: list[NexusCompletion] = completions
        super().__init__(well_name=well_name, completions=completions, unit_system=unit_system)

    def __repr__(self) -> str:
        return generic_repr(self)

    def __str__(self) -> str:
        return generic_str(self)

    @property
    def perforations(self) -> Sequence[NexusCompletion]:
        """Returns a list of all of the perforations for the well."""

        activations = filter(NexusCompletion.completion_is_perforation, self.__completions)
        return list(activations)

    @property
    def first_perforation(self) -> Optional[NexusCompletion]:
        """Returns the first perforation for the well."""
        if len(self.perforations) == 0:
            return None

        return self.perforations[0]

    @property
    def shutins(self) -> Sequence[NexusCompletion]:
        """Returns a list of all of the shut-ins for the well."""

        shutins = filter(NexusCompletion.completion_is_shutin, self.__completions)
        return list(shutins)

    @property
    def last_shutin(self) -> Optional[NexusCompletion]:
        """Returns the last shut-in for the well in the Wellspec file."""
        if len(self.shutins) == 0:
            return None

        return self.shutins[-1]

    @property
    def printable_well_info(self) -> str:
        """Returns some printable well information in string format."""
        printable_dates_of_completions = ", ".join(self.dates_of_completions)
        well_info = \
            f"""
    Well Name: {self.well_name}
    First Perforation: {'N/A' if self.first_perforation is None else self.first_perforation.date}
    Last Shut-in: {'N/A' if self.last_shutin is None else self.last_shutin.date}
    Dates Changed: {'N/A' if len(self.dates_of_completions) == 0 else printable_dates_of_completions}
    """

        return well_info

    @property
    def completion_events(self) -> list[tuple[str, Union[int, tuple[float, float]]]]:
        """Returns a list of dates and values representing either the layer, or the depths of each perforation."""
        events = []
        using_k_values: Optional[bool] = None

        for completion in self.__completions:
            is_perforation = NexusCompletion.completion_is_perforation(completion)
            if not is_perforation:
                continue
            if completion.k is not None and using_k_values is not False:
                using_k_values = True
                events.append((completion.date, completion.k))
            elif completion.depth_to_top is not None and using_k_values is not True:
                using_k_values = False
                events.append((completion.date, (completion.depth_to_top, completion.depth_to_bottom)))

        return events

    def find_completions(self, completion_properties:  dict[str, None | float | int | str] | NexusCompletion) -> \
            list[NexusCompletion]:
        """Returns a list of all completions that match the completion properties provided.

        Args:
        ----
            completion_properties (dict | NexusCompletion): keys as the attributes and values as the value to match \
            from the well.

        Returns:
        -------
            list[NexusCompletion] that match the completion properties provided
        """
        matching_completions = []
        if isinstance(completion_properties, NexusCompletion):
            perf_props = completion_properties.to_dict(add_units=False)
        else:
            perf_props = completion_properties
        perf_props_without_nones = {k: v for k, v in perf_props.items() if v is not None}
        for i, perf in enumerate(self.__completions):
            for prop, value in perf_props_without_nones.items():
                if getattr(perf, prop) == value:
                    # go to the next perf if a value from the dictionary doesn't match
                    continue
                else:
                    break
            else:
                # if all the conditions match then append the perf to completions
                matching_completions.append(perf)
        return matching_completions

    def find_completion(self,
                        completion_properties: dict[str, None | float | int | str] | NexusCompletion) \
            -> NexusCompletion:
        """Returns precisely one completion which matches all the properties provided.
            If several completions match it raise a ValueError.

        Args:
        ----
            completion_properties (dict | NexusCompletion):

        Returns:
        -------
            NexusCompletion that matches all completion properties provided.
        """
        matching_completions = self.find_completions(completion_properties)
        if len(matching_completions) != 1:
            raise ValueError(f'Could not find single unique completion that matches property provided. Instead found: '
                             f'{len(matching_completions)} completions')
        return matching_completions[0]

    def get_completion_by_id(self, id: UUID) -> NexusCompletion:
        """Returns the completion that matches the id provided."""
        for completion in self.__completions:
            if completion.id == id:
                return completion
        raise ValueError('No completion found for id: {id}')

    def _add_completion_to_memory(self, date: str, completion_properties: dict[str, None | float | int | str],
                                  date_format: DateFormat, completion_index: Optional[int] = None) -> NexusCompletion:
        """Adds a perforation with the properties specified in completion_properties_list,
            if index is none then adds it to the end of the perforation list.

        Args:
        ----
            date (str): date at which the perforation should be added
            completion_properties (dict[str, str | float | int]):
            completion_index (Optional[int]):
        """
        completion_properties['date'] = date
        completion_properties['unit_system'] = self.unit_system
        new_completion = NexusCompletion.from_dict(completion_properties, date_format)
        if completion_index is None:
            completion_index = len(self.__completions)
        self.__completions.insert(completion_index, new_completion)
        return new_completion

    def _remove_completion_from_memory(self, completion_to_remove: NexusCompletion | UUID) -> None:
        if isinstance(completion_to_remove, NexusCompletion):
            completion_to_remove = self.find_completion(completion_to_remove)
            completion_to_remove = completion_to_remove.id
        completion_index_to_remove = [x.id for x in self.__completions].index(completion_to_remove)
        self.__completions.pop(completion_index_to_remove)

    def _modify_completion_in_memory(self, new_completion_properties: dict[str, Union[None, float, int, str]],
                                     completion_to_modify: NexusCompletion | UUID,
                                     ) -> None:
        if isinstance(completion_to_modify, NexusCompletion):
            completion = self.find_completion(completion_to_modify)
        else:
            completion = self.get_completion_by_id(completion_to_modify)
        completion.update(new_completion_properties)

    def _modify_completions_in_memory(self, new_completion_properties: dict[str, Union[None, float, int, str]],
                                      completions_to_modify: list[NexusCompletion | UUID]) -> None:
        for completion in completions_to_modify:
            if isinstance(completion, UUID):
                modify_this_completion = self.get_completion_by_id(completion)
                modify_this_completion.update(new_completion_properties)
            else:
                completion.update(new_completion_properties)

    def _remove_completions_from_memory(self, completions_to_remove: Sequence[NexusCompletion | UUID]) -> None:
        # TODO improve comparison of dates with datetime libs
        """Removes completions from a list of completions or completion IDs
        Args:
            completions_to_remove (list[NexusCompletion | UUID]): Completions to be removed.
        """
        for completion in completions_to_remove:
            self._remove_completion_from_memory(completion)
